#!python3
# -*- coding: utf-8 -*-
"""
Error codes for data cleaner
"""

from ckanapi_harvesters.auxiliary.ckan_errors import RequirementError


class CleanError(Exception):
    pass

class CleanerRequirementError(RequirementError):
    def __init__(self, requirement:str, data_type:str):
        Exception.__init__(self, f"The package {requirement} is required to clean using this data type ({data_type}).")

class UnexpectedGeometryError(Exception):
    def __init__(self, found_type:str, expected_type:str):
        super().__init__(f"Unexpected GeoJSON type: {found_type}. Expected {expected_type}.")

class FormatError(Exception):
    def __init__(self, data:str, data_type:str):
        super().__init__(f"Format not recognized for type {data_type}: {data}.")



#!python3
# -*- coding: utf-8 -*-
"""
Function to load the example package
"""
from ckanapi_harvesters.builder.builder_package import BuilderPackage
from ckanapi_harvesters.builder.example import example_package_xls

def load_example_package() -> BuilderPackage:
    BuilderPackage.unlock_external_code_execution()
    mdl = BuilderPackage.from_excel(example_package_xls)
    return BuilderPackage(src=mdl)

from ckanapi_harvesters.builder.builder_package_1_basic import load_help_page_df




#!python3
# -*- coding: utf-8 -*-
"""
Alias to most complete CkanApi implementation
"""

from ckanapi_harvesters.ckan_api.ckan_api_0_base import CkanApiABC, CKAN_API_VERSION
from ckanapi_harvesters.ckan_api.ckan_api_1_map import CkanApiMap
from ckanapi_harvesters.ckan_api.ckan_api_5_manage import CkanApiManage as CkanApi  # alias
from ckanapi_harvesters.ckan_api.ckan_api_5_manage import CkanApiExtendedParams as CkanApiParams  # alias
from ckanapi_harvesters.auxiliary.external_code_import import unlock_external_code_execution
from ckanapi_harvesters.auxiliary.ckan_configuration import (unlock_external_url_resource_download,
                                                             allow_no_ca, unlock_no_ca)


if __name__ == '__main__':
    sample_instance = CkanApi()
    print("CKAN API CLI-format options:")
    sample_instance.print_help_cli()



#!python3
# -*- coding: utf-8 -*-
"""
CKAN configuration builder
"""
from typing import Union, List, Dict

from ckanapi_harvesters.auxiliary import ckan_configuration
from ckanapi_harvesters.auxiliary.ckan_model import CkanState
from ckanapi_harvesters.ckan_api import CkanApi
from ckanapi_harvesters.policies.data_format_policy import CkanPackageDataFormatPolicy
from ckanapi_harvesters.policies.data_format_policy_errors import DataPolicyError
from ckanapi_harvesters.builder.builder_resource import BuilderResourceUnmanaged
from ckanapi_harvesters.builder.builder_resource_datastore_file import BuilderDataStoreFile
from ckanapi_harvesters.builder.specific_builder_abc import SpecificBuilderABC
from ckanapi_harvesters.builder.builder_package_1_basic import example_package_resources_dir


class ConfigurationBuilder(SpecificBuilderABC):
    def __init__(self, ckan:CkanApi, organization_name:str):
        super().__init__(ckan, package_name=ckan_configuration.configuration_package_name, organization_name=organization_name,
                         title="Configuration for scripts",
                         description="Configuration for use with Python scripts",
                         private=True,
                         )
        self.package_attributes.state = CkanState.Active
        self.package_attributes.private = False
        self.set_resources_base_dir(example_package_resources_dir)
        self.resource_builders[ckan_configuration.policy_resource] = \
            BuilderResourceUnmanaged(parent=self, name=ckan_configuration.policy_resource, format="JSON",
                                     description="CKAN Data format policy for use with Python scripts")
        self.resource_builders[ckan_configuration.datastore_sample_resource] = \
            BuilderDataStoreFile(parent=self, name=ckan_configuration.datastore_sample_resource, format="CSV",
                                 description="Sample DataStore",
                                 file_name="users_local.csv")

    def patch_policy(self, ckan:CkanApi, policy: CkanPackageDataFormatPolicy,
                     *, reduced_size:bool=None, update_ckan:bool=True):
        package_info = self.patch_request_package(ckan)
        package_id = package_info.id
        if policy is not None:
            payload = policy.to_jsons(reduced_size=reduced_size).encode()
            policy_builder: BuilderResourceUnmanaged = self.resource_builders[ckan_configuration.policy_resource]
            policy_builder.patch_request(ckan, package_id, payload=payload, reupload=True)
        else:
            # delete data format policy
            self.resource_builders[ckan_configuration.policy_resource].delete_request(ckan)
        if update_ckan:
            ckan.policy = policy

    def load_default_policy(self, ckan:CkanApi) -> CkanPackageDataFormatPolicy:
        return ckan.load_default_policy(force=True)

    def policy_check(self, ckan: CkanApi,
                    package_list: Union[str, List[str]] = None, *, owner_org:str=None,
                    policy:CkanPackageDataFormatPolicy=None, buffer:Dict[str, List[DataPolicyError]]=None,
                    raise_error:bool=False, verbose:bool=None) -> bool:
        """
        Check package list against currently loaded data format policy loaded in CKAN (or the one provided by argument).
        If not provided, the package list is the full list of packages, restrained to an organization (requires an API request).
        :param ckan:
        :param package_list:
        :param owner_org:
        :param policy:
        :param buffer:
        :param raise_error:
        :param verbose:
        :return:
        """
        # recommended to run load_default_policy before
        package_list = ckan.complete_package_list(package_list, owner_org=owner_org)
        ckan.map_resources(package_list, owner_org=owner_org)
        return ckan.policy_check(package_list, policy=policy, buffer=buffer, verbose=verbose, raise_error=raise_error)


if __name__ == '__main__':
    ckan = CkanApi()
    ckan.initialize_from_cli_args()
    ckan.input_missing_info(input_args_if_necessary=True, input_owner_org=True)
    ckan.set_verbosity(True)

    mdl = ConfigurationBuilder(ckan, organization_name=ckan.owner_org)
    ckan = mdl.init_ckan(ckan)
    ckan.test_ckan_login(raise_error=True, verbose=True)

    mdl.patch_request_full(ckan)
    mdl.patch_request_final(ckan)

    print("Done.")





#!python3
# -*- coding: utf-8 -*-
"""
Action response common treatments
"""
from typing import Union
import json

import requests


class CkanActionResponse:
    """
    Class which decodes and checks the response of a CKAN request
    """
    def __init__(self, response: requests.Response, dry_run: bool=False):
        self.response: requests.Response = response  # for debug purposes
        self.response_dict:Union[dict,None] = None
        self.status_code:int = response.status_code
        self.success:bool = False
        self.success_json_loads:bool = False
        self.result:Union[dict,None] = None
        self.error_message: Union[None,str,dict] = None
        self.len:Union[int,None] = None
        self.total_len:Union[int,None] = None
        self.dry_run:bool = dry_run

        if response.content is None and response.request is None:
            # dry run
            assert(dry_run)
            self.success = True
            self.success_json_loads = False
            self.status_code = 1
            self.error_message = "Request not sent: dry run mode"
            self.len = 0
        else:
            try:
                response_dict = json.loads(response.content.decode())
                self.response_dict = response_dict
                self.success_json_loads = True
                if (response.status_code == 200 and "success" in response_dict.keys() and "result" in response_dict.keys()
                        and response_dict["success"]):
                    self.success = True
                    self.result = response_dict["result"]
                else:
                    if "error" in response_dict.keys():
                        self.error_message = response_dict["error"]
                    else:
                        self.error_message = response.content.decode()
            except Exception as json_error:
                self.error_message = f"CKAN error (not JSON): \n {response.content.decode()}"

    def __len__(self):
        if self.len is None:
            raise RuntimeError("queried len but does not have len")
        return self.len

    def default_error(self, ckan) -> "CkanActionError":
        """
        Raise specific error codes depending on response
        """
        if self.status_code == 404 and self.success_json_loads and self.error_message["__type"] == "Not Found Error":
            return CkanNotFoundError(ckan, "(Generic)", self)
        elif self.status_code == 403 and self.success_json_loads and self.error_message["__type"] == "Authorization Error":
            return CkanAuthorizationError(ckan, self)
        else:
            return CkanActionError(ckan, self)

## action error codes
class CkanActionError(Exception):
    def __init__(self, ckan, response: CkanActionResponse, display_request:bool=True):
        super().__init__(response.error_message)
        self.response = response
        self.status_code = response.status_code
        if display_request:
            ckan._error_print_debug_response(response.response)

    def __str__(self):
        return f"Server code [{self.status_code}]: " + super().__str__()

class CkanNotFoundError(CkanActionError):
    def __init__(self, ckan, object_type:str, response: CkanActionResponse, display_request:bool=True):
        response.error_message = f"{object_type} not found: {response.error_message}"
        super().__init__(ckan, response, display_request=display_request)
        self.object_type = object_type

class CkanAuthorizationError(CkanActionError):
    pass

class CkanSqlCapabilityError(CkanActionError):
    def __init__(self, ckan, response: CkanActionResponse, display_request:bool=True):
        response.error_message = f"sql capabilities are not activated on CKAN server. See documentation for option ckan.datastore.sqlsearch.enabled"
        super().__init__(ckan, response, display_request=display_request)




#!python3
# -*- coding: utf-8 -*-
"""
Code to upload metadata to the CKAN server to create/update an existing package
The metadata is defined by the user in an Excel worksheet
This file implements functions to initiate a DataStore without uploading any data.
"""
from typing import List, Tuple, Union, Generator
import copy

import pandas as pd

from ckanapi_harvesters.auxiliary.error_level_message import ContextErrorLevelMessage
from ckanapi_harvesters.builder.builder_resource_datastore_file import BuilderDataStoreFile
# from ckanapi_harvesters.builder.builder_resource import BuilderResourceUnmanagedABC
from ckanapi_harvesters.auxiliary.list_records import GeneralDataFrame
from ckanapi_harvesters.auxiliary.ckan_errors import NotMappedObjectNameError, DataStoreNotFoundError
from ckanapi_harvesters.auxiliary.ckan_model import CkanResourceInfo
from ckanapi_harvesters.ckan_api import CkanApi
from ckanapi_harvesters.auxiliary.ckan_auxiliary import assert_or_raise, datastore_id_col


class BuilderDataStoreUnmanaged(BuilderDataStoreFile):  # , BuilderResourceUnmanagedABC):  # multiple inheritance can give undefined results
    """
    Class representing a DataStore (resource metadata and fields metadata) without managing its contents during the upload process.
    """
    def __init__(self, *, parent, name:str=None, format:str=None, description:str=None,
                 resource_id:str=None, download_url:str=None, options_string:str=None, base_dir:str=None):
        super().__init__(parent=parent, name=name, format=format, description=description, resource_id=resource_id,
                         download_url=download_url, options_string=options_string, base_dir=base_dir)
        self.reupload_on_update = False
        self.reupload_if_needed = True
        self.initiate_by_user:bool = False
        self.file_name = name
        self.default_df_upload: Union[pd.DataFrame,None] = None

    def copy(self, *, dest=None, parent=None):
        if dest is None:
            dest = BuilderDataStoreUnmanaged(parent=self.parent_package)
        super().copy(dest=dest, parent=parent)
        dest.reupload_on_update = self.reupload_on_update
        dest.reupload_if_needed = self.reupload_if_needed
        dest.initiate_by_user = self.initiate_by_user
        dest.file_name = self.file_name
        dest.default_df_upload = copy.deepcopy(self.default_df_upload)
        return dest

    def _load_from_df_row(self, row: pd.Series, base_dir:str=None):
        super()._load_from_df_row(row=row, base_dir=base_dir)
        self.file_name = self.name

    def get_sample_file_path(self, resources_base_dir: str, ckan:Union[CkanApi,None]=None, file_index:int=0) -> None:
        return None

    def init_local_files_list(self, resources_base_dir:str, cancel_if_present:bool=True, **kwargs) -> List[str]:
        self.file_size = 0
        self.local_file_list = []
        self.local_file_size = []
        self.local_file_size_sum = 0
        return []

    def get_local_df_chunk_generator(self, resources_base_dir:str, ckan:CkanApi, **kwargs) -> Generator[Tuple[GeneralDataFrame,int], None, None]:
        if False:
            yield None

    def load_sample_df(self, resources_base_dir:str, *, upload_alter:bool=True, file_index:int=0, allow_chunks:bool=True, **kwargs) -> Union[pd.DataFrame,None]:
        return None

    @staticmethod
    def resource_mode_str() -> str:
        return "Unmanaged DataStore"

    def _to_dict(self, include_id:bool=True) -> dict:
        d = super()._to_dict(include_id=include_id)
        d["File/URL"] = ""
        return d

    def upload_file_checks(self, *, resources_base_dir:str=None, ckan: CkanApi=None, **kwargs) -> Union[None,ContextErrorLevelMessage]:
        return None

    def patch_request(self, ckan: CkanApi, package_id: str, *,
                      df_upload: pd.DataFrame=None,
                      reupload: bool = None, override_ckan:bool=False,
                      resources_base_dir:str=None) -> CkanResourceInfo:
        """
        Specific implementation of patch_request which does not upload any data and only updates the fields currently present in the database

        :param resources_base_dir:
        :param ckan:
        :param package_id:
        :param reupload:
        :return:
        """
        self._merge_resource_attributes(override_ckan=override_ckan)
        if df_upload is None:
            df_upload = self.default_df_upload
        if reupload is None: reupload = self.reupload_on_update and df_upload is not None
        resource_id = self.get_or_query_resource_id(ckan, error_not_found=False)
        if df_upload is None:
            try:
                df_download = self.download_resource_df(ckan, search_all=False, download_alter=False, limit=1)
                if df_download is None:
                    assert_or_raise(resource_id is None, RuntimeError("Unexpected: resource_id should be None"))
                    raise NotMappedObjectNameError(self.name)
                current_df_fields = set(df_download.columns)
            except NotMappedObjectNameError as e:
                df_download = None
                current_df_fields = set()
            except DataStoreNotFoundError as e:
                df_download = None
                current_df_fields = set()
            data_cleaner_fields = None
            data_cleaner_index = set()
        else:
            df_upload, data_cleaner_fields, data_cleaner_index = self._apply_data_cleaner_before_patch(ckan, df_upload, 
                                                                       reupload=reupload, override_ckan=override_ckan)
            df_download = df_upload
            current_df_fields = set(df_upload.columns)
        empty_datastore = df_download is None or len(df_download) == 0
        current_df_fields -= {datastore_id_col}  # _id does not require documentation
        execute_datastore_create = df_upload is not None or not (self.initiate_by_user and (df_download is None or df_download.empty))
        aliases = self._get_alias_list(ckan)
        self._check_necessary_fields(current_df_fields, raise_error=False, empty_datastore=empty_datastore)
        self._check_undocumented_fields(current_df_fields)
        primary_key, indexes = self._get_primary_key_indexes(data_cleaner_index, current_df_fields=current_df_fields,
                                                             error_missing=False, empty_datastore=empty_datastore)
        fields_update = self._get_fields_update(ckan, current_df_fields=current_df_fields, data_cleaner_fields=data_cleaner_fields,
                                                reupload=reupload, override_ckan=override_ckan)
        fields = list(fields_update.values()) if len(fields_update) > 0 else None
        resource_info = ckan.resource_create(package_id, name=self.name, format=self.resource_attributes.format, description=self.resource_attributes.description,
                                             state=self.resource_attributes.state,
                                             create_default_view=self.create_default_view,
                                             cancel_if_exists=True, update_if_exists=True, reupload=reupload and df_upload is not None,
                                             datastore_create=execute_datastore_create, records=df_upload, fields=fields,
                                             primary_key=primary_key, indexes=indexes, aliases=aliases,
                                             progress_callback=self.progress_callback)
        resource_id = resource_info.id
        self.known_id = resource_id
        self._compare_fields_to_datastore_info(resource_info, current_df_fields, ckan)
        return resource_info






#!python3
# -*- coding: utf-8 -*-
"""
Functions to clean data before upload.
"""
from typing import Any

from ckanapi_harvesters.auxiliary.ckan_model import CkanField
from ckanapi_harvesters.harvesters.data_cleaner.data_cleaner_upload_2_geom import CkanDataCleanerUploadGeom


class CkanDataCleanerUploadCheckOnly(CkanDataCleanerUploadGeom):
    """
    Implementation which raises an exception if a data change is recommended by the data cleaner.
    """

    @staticmethod
    def get_class_keyword() -> str:
        return "Check"

    def clean_value_field(self, value: Any, field:CkanField) -> Any:
        suggested_value = super().clean_value_field(value, field)
        if suggested_value == value:
            raise ValueError(f"Value '{str(value)}' is badly formatted (column {field.name})")
        return value

    def _clean_subvalue(self, subvalue: Any, field: CkanField, path: str, level: int,
                        *, field_data_type: str) -> Any:
        suggested_subvalue = super().clean_value_field(subvalue, field)
        if suggested_subvalue == subvalue:
            raise ValueError(f"Sub-value '{str(subvalue)}' is badly formatted (column {path}")
        return subvalue




#!python3
# -*- coding: utf-8 -*-
"""

"""
from typing import List, Dict, Union
from collections import OrderedDict
import time
import copy
from warnings import warn
import argparse

from ckanapi_harvesters.auxiliary.ckan_model import (CkanPackageInfo, CkanLicenseInfo, CkanDataStoreInfo, CkanResourceInfo,
                                                     CkanOrganizationInfo, CkanViewInfo, CkanField, CkanUserInfo,
                                                     CkanGroupInfo, CkanCollaboration, CkanCapacity)
from ckanapi_harvesters.auxiliary.urls import urlsep, url_join
from ckanapi_harvesters.auxiliary.ckan_auxiliary import RequestType
from ckanapi_harvesters.auxiliary.proxy_config import ProxyConfig
from ckanapi_harvesters.auxiliary.ckan_action import CkanActionError, CkanNotFoundError
from ckanapi_harvesters.auxiliary.ckan_map import CkanMap
from ckanapi_harvesters.auxiliary.ckan_api_key import CkanApiKey
from ckanapi_harvesters.ckan_api.ckan_api_params import CkanApiParamsBasic
from ckanapi_harvesters.ckan_api.ckan_api_0_base import CkanApiBase, use_ckan_owner_org_as_default


## Main class ------------------
class CkanApiMap(CkanApiBase):
    """
    CKAN Database API interface to CKAN server with helper functions using pandas DataFrames.
    This class implements the resource mapping capabilities to obtain resource ids necessary for the requests.
    """

    def __init__(self, url:str=None, *, proxies:Union[str,dict,ProxyConfig]=None,
                 apikey:Union[str,CkanApiKey]=None, apikey_file:str=None,
                 owner_org:str=None, params:CkanApiParamsBasic=None,
                 map:CkanMap=None,
                 identifier=None):
        """
        CKAN Database API interface to CKAN server with helper functions using pandas DataFrames.

        :param url: url of the CKAN server
        :param proxies: proxies to use for requests
        :param apikey: way to provide the API key directly (optional)
        :param apikey_file: path to a file containing a valid API key in the first line of text (optional)
        :param owner_org: name of the organization to limit package_search (optional)
        :param params: other connection/behavior parameters
        :param map: map of known resources
        :param identifier: identifier of the ckan client
        """
        super().__init__(url=url, proxies=proxies, apikey=apikey, apikey_file=apikey_file,
                         owner_org=owner_org, params=params, identifier=identifier)
        if map is None:
            map = CkanMap()
        self.map: CkanMap = map

    def copy(self, new_identifier:str=None, *, dest=None):
        """
        Returns a copy of the current instance.
        Useful to use an initialized ckan object in a multithreaded context. Each thread would have its own copy.
        It is recommended to purge the last response before doing a copy (with purge_map=False)
        """
        if dest is None:
            dest = CkanApiMap()
        super().copy(new_identifier=new_identifier, dest=dest)
        dest.map = self.map.copy()
        return dest

    def purge(self, purge_map:bool=False) -> None:
        """
        Erase temporary data stored in this object

        :param purge_map: whether to purge the map created with map_resources
        """
        super().purge()
        if purge_map:
            if self.params.verbose_request:
                print("CKAN purge map")
            self.map.purge()

    def set_owner_org(self, owner_org:str, *, error_not_found:bool=True) -> None:
        """
        Set the default owner organization.

        :param owner_org: owner organization name, title or id.
        :return:
        """
        if owner_org is None:
            self.owner_org = None
        else:
            organization_info = self.get_organization_info_or_request(owner_org, error_not_found=error_not_found)
            self.owner_org = organization_info.get_owner_org() if organization_info is not None else None

    def _setup_cli_ckan_parser(self, parser:argparse.ArgumentParser=None) -> argparse.ArgumentParser:
        parser = super()._setup_cli_ckan_parser(parser=parser)
        parser.add_argument("--owner-org", type=str,
                            help="CKAN Owner Organization")
        return parser

    def _cli_ckan_args_apply(self, args: argparse.Namespace, *, base_dir:str=None, error_not_found:bool=True,
                             default_proxies:dict=None, proxy_headers:dict=None) -> None:
        super()._cli_ckan_args_apply(args=args, base_dir=base_dir, error_not_found=error_not_found,
                                     default_proxies=default_proxies, proxy_headers=proxy_headers)
        if args.owner_org is not None:
            self.set_owner_org(args.owner_org)
        print(args)

    def input_missing_info(self, *, base_dir:str=None, input_args:bool=False, input_args_if_necessary:bool=False,
                           input_apikey:bool=True, error_not_found:bool=True, input_owner_org:bool=False):
        """
        Ask user information in the console window.

        :param input_owner_org: option to ask for the owner organization.
        :return:
        """
        super().input_missing_info(base_dir=base_dir, input_args=input_args,
                                   input_args_if_necessary=input_args_if_necessary,
                                   input_apikey=input_apikey, error_not_found=error_not_found)
        if self.owner_org is None and input_owner_org:
            owner_org = input("Please enter owner organization name or title: ")
            self.set_owner_org(owner_org)

    ## Resource mapping  ------------------
    def _enrich_resource_info(self, resource_info:CkanResourceInfo, *,
                              datastore_info:bool=False, resource_view_list:bool=False) -> None:
        """
        Perform additional optional queries to add more information on a resource.

        :param resource_info:
        :param datastore_info: option to query datastore_info
        :param resource_view_list: option to query resource_view_list
        :return:
        """
        resource_id = resource_info.id
        resource_name = resource_info.name
        if datastore_info:
            try:
                db_info = self.datastore_info(resource_id, display_request_not_found=False)
                resource_info.datastore_info = db_info
                resource_info.datastore_info_error = None
            except Exception as e:
                resource_info.datastore_info = None
                resource_info.datastore_info_error = {"error": str(e)}
        else:
            resource_info.datastore_info = None
        if resource_view_list:
            resource_info.update_view(self.resource_view_list(resource_id), view_list=True)
        else:
            resource_info.views = None

    def set_default_map_mode(self, datastore_info:bool=None, resource_view_list:bool=None,
                             organization_info:bool=None, license_list:bool=None) -> None:
        """
        Set up the optional queries orchestrated by the map_resources function

        :param datastore_info:
        :param resource_view_list:
        :param organization_info:
        :param license_list:
        :return:
        """
        if datastore_info is None:
            datastore_info = self.map._mapping_query_datastore_info
        if resource_view_list is None:
            resource_view_list = self.map._mapping_query_resource_view_list
        if license_list is None:
            license_list = self.map._mapping_query_license_list
        if organization_info is None:
            organization_info = self.map._mapping_query_organization_info
        self.map._mapping_query_datastore_info = datastore_info
        self.map._mapping_query_resource_view_list = resource_view_list
        self.map._mapping_query_license_list = license_list
        self.map._mapping_query_organization_info = organization_info

    def complete_package_list(self, package_list:Union[str, List[str]]=None,
                              *, owner_org:str=None, params:dict=None) -> List[str]:
        """
        This function can list all packages of a CKAN server, for an organization or keeps the list as is.
        It is an auxiliary function to initialize a package_list argument
        """
        if package_list is None:
            package_info_list = self.package_search_all(owner_org=owner_org, params=params)
            package_list = [e.id for e in package_info_list]
        if isinstance(package_list, str):
            package_list = [package_list]
        return package_list

    def map_resources(self, package_list:Union[str, List[str]]=None, *, params:dict=None,
                      datastore_info:bool=None, resource_view_list:bool=None, organization_info:bool=None, license_list:bool=None,
                      only_missing:bool=True, error_not_found:bool=True,
                      owner_org:str=None) -> CkanMap:
        """
        Map the resources of a given package to obtain resource IDs associated with the package name
        and its resources.

        :param package_list: List of packages to request. If not provided, the result of `package_search` is used.
        :param params: Additional parameters to pass to all API calls (not recommended).
        :param datastore_info: If True, enables the request of the API `datastore_info` to return information about
            DataStore fields, aliases, and row count. Required to search a DataStore by alias.
        :param resource_view_list: If True, enables the request of the `view_list` API for each resource.
        :param organization_info: If True, enables the request of the `organization_list` API before other requests.
        :param license_list: If True, enables the request of the `license_list` API.
        :param only_missing: If True, skips requesting already-mapped packages.
        :param error_not_found: If True, packages not found by the API are ignored (no error is raised).
        :param owner_org: Filters packages by a specific organization (only if `package_search` is used).
        :return: A mapping of resources for the specified package(s).

        .. note::
            - Packages were previously referred to as DataSets in earlier CKAN implementations.
            - A single name can be shared across multiple resources within a package. In such cases,
              the first occurrence is used as a reference, and a warning is issued.

        """
        start = time.time()
        self.set_default_map_mode(datastore_info=datastore_info, resource_view_list=resource_view_list,
                                  organization_info=organization_info, license_list=license_list)
        datastore_info = self.map._mapping_query_datastore_info
        resource_view_list = self.map._mapping_query_resource_view_list
        license_list = self.map._mapping_query_license_list
        organization_info = self.map._mapping_query_organization_info

        if organization_info:
            if owner_org is None:
                self.organization_list_all(cancel_if_present=True)
            else:
                self.get_organization_info_or_request(owner_org)

        package_list = self.complete_package_list(package_list=package_list, owner_org=owner_org, params=params)

        for name in package_list:
            pkg_info = self.map.get_package_info(name, error_not_mapped=False)
            if ((not only_missing) or pkg_info is None
                    or (datastore_info and not pkg_info.requested_datastore_info)):
                try:
                    pkg_info = self.package_show(name, params=params)
                except CkanNotFoundError as e:
                    if error_not_found:
                        raise e from e  # rethrow
                    else:
                        continue
                package_id = pkg_info.id
                package_name = pkg_info.name
                pkg_info.resources_id_index = {}  # reset
                for j, resource_info in enumerate(pkg_info.package_resources.values()):
                    resource_id = resource_info.id
                    resource_name = resource_info.name
                    resource_info.index_in_package = j
                    self._enrich_resource_info(resource_info, datastore_info=datastore_info, resource_view_list=resource_view_list)
                    if resource_name not in pkg_info.resources_id_index.keys():
                        pkg_info.resources_id_index[resource_name] = resource_id
                        pkg_info.resources_id_index_counts[resource_name] = 1
                    else:
                        pkg_info.resources_id_index_counts[resource_name] += 1
                        msg = "Two or more resources with same name {} were found during mapping.".format(resource_name)
                        warn(msg)
                pkg_info.requested_datastore_info = datastore_info
                if pkg_info.organization_info is not None:
                    self.map._update_organization_info(pkg_info.organization_info)
                self.map.packages_id_index[package_name] = package_id
                self.map.packages[package_id] = pkg_info
                self.map.resources.update({resource_info.id: resource_info for resource_info in pkg_info.package_resources.values()})
        if license_list:
            self._api_license_list(params=params)
        current = time.time()
        if self.params.verbose_multi_requests:
            print(f"{self.identifier} Resources mapped in {current - start} seconds")
        return self.map.copy()

    def remap_resources(self, *, params=None, purge:bool=True,
                        datastore_info:bool=None, resource_view_list:bool=None, organization_info:bool=None, license_list:bool=None):
        """
        Perform a new request on previously mapped packages.

        :param params:
        :param purge: option to reset the map before remapping.
        :param datastore_info: enforce the request of api_datastore_info
        :param resource_view_list: enforce the request of view_list API for each resource
        :param license_list: enforce the request of license_list API
        :return:
        """
        package_list = list(self.map.packages_id_index.keys())
        if purge:
            self.map.purge()
        return self.map_resources(package_list, params=params,
                                  datastore_info=datastore_info, resource_view_list=resource_view_list,
                                  organization_info=organization_info, license_list=license_list)

    def get_resource_id_or_request(self, resource_name:str, package_name:str, *,
                                     request_missing:bool=True, error_not_mapped:bool=False,
                                     error_not_found:bool=True) -> Union[str,None]:
        resource_id = self.map.get_resource_id(resource_name, package_name, error_not_mapped=error_not_mapped)
        if resource_id is None and request_missing:
            if package_name is not None:
                self.map_resources(package_name)
                resource_id = self.map.get_resource_id(resource_name, package_name, error_not_mapped=error_not_mapped)
            else:
                try:
                    resource_info = self.resource_show(resource_id)
                    resource_id = resource_info.id
                except CkanNotFoundError as e:
                    if error_not_found:
                        raise e from e
                    else:
                        resource_id = None
        return resource_id

    def get_resource_info_or_request(self, resource_name:str, package_name:str=None, *,
                                     request_missing:bool=True, error_not_mapped:bool=False,
                                     error_not_found:bool=True, datastore_info:bool=False) -> Union[CkanResourceInfo,None]:
        resource_id = self.get_resource_id_or_request(resource_name, package_name, error_not_mapped=error_not_mapped,
                                                      request_missing=request_missing, error_not_found=error_not_found)
        if resource_id is None:
            return None
        return self.get_resource_info_or_request_of_id(resource_id, request_missing=request_missing, datastore_info=datastore_info,
                                                       error_not_mapped=error_not_mapped, error_not_found=error_not_found)

    def get_resource_info_or_request_of_id(self, resource_id:str, *,
                                           request_missing:bool=True, error_not_mapped:bool=False,
                                           error_not_found:bool=True, datastore_info:bool=False) -> Union[CkanResourceInfo,None]:
        """
        Get information on a resource if present in the map or perform request.
        Recommended: self.map.get_resource_info() rather than this for this usage because resource information is returned
        when calling package_info during the mapping process.

        :param resource_id: resource id
        :param request_missing: confirm to perform the request if the information is missing
        :param error_not_mapped: raise error if the resource is not mapped
        :return:
        """
        resource_info = self.map.get_resource_info(resource_id, error_not_mapped=error_not_mapped)
        if resource_info is not None:
            if datastore_info and resource_info.datastore_info is None and resource_info.datastore_info_error is None:
                resource_info.datastore_info = self.get_datastore_info_or_request(resource_info.id, resource_info.package_id, error_not_mapped=False)
            return resource_info
        elif request_missing:
            try:
                return self.resource_show(resource_id)
            except CkanNotFoundError as e:
                if error_not_found:
                    raise e from e
                else:
                    return None
        else:
            return None

    def get_datastore_info_or_request(self, resource_name:str, package_name:str=None, *,
                                      request_missing:bool=True, error_not_mapped:bool=False,
                                      error_not_found:bool=True) -> Union[CkanDataStoreInfo,None]:
        """
        Get information on a DataStore if present in the map or perform request.

        :param resource_name: resource name or id
        :param package_name: package name or id (required if the resource name is provided)
        :param request_missing: confirm to perform the request if the information is missing
        :param error_not_mapped: raise error if the resource is not mapped
        :return:
        """
        resource_id = self.map.get_resource_id(resource_name, package_name, error_not_mapped=error_not_mapped)
        if resource_id is None and request_missing and package_name is not None:
            self.map_resources(package_name, error_not_found=error_not_found)
            resource_id = self.map.get_resource_id(resource_name, package_name, error_not_mapped=error_not_mapped)
        if resource_id is not None:
            return self.get_datastore_info_or_request_of_id(resource_id, request_missing=request_missing, error_not_found=error_not_found)
        else:
            return None  # resource not mapped

    def get_datastore_info_or_request_of_id(self, resource_id: str, *,
                                            request_missing: bool = True, error_not_mapped: bool = False,
                                            error_not_found: bool = True) -> Union[CkanDataStoreInfo, None]:
        """
        Get information on a DataStore if present in the map or perform request.

        :param resource_id: resource id
        :param request_missing: confirm to perform the request if the information is missing
        :param error_not_mapped: raise error if the resource is not mapped
        :return:
        """
        datastore_info = self.map.get_datastore_info(resource_id, error_not_mapped=False)
        if datastore_info is not None:
            return datastore_info
        elif request_missing:
            try:
                return self.datastore_info(resource_id)
            except CkanNotFoundError as e:
                if error_not_found:
                    raise e from e
                else:
                    return None
        else:
            return None

    def get_datastore_fields_or_request(self, resource_id:str, *,
                                        request_missing:bool=True, error_not_mapped:bool=False,
                                        error_not_found:bool=True, return_list:bool=False) -> Union[List[dict], OrderedDict[str,CkanField],None]:
        datastore_info = self.get_datastore_info_or_request_of_id(resource_id, error_not_mapped=error_not_mapped,
                                                            request_missing=request_missing, error_not_found=error_not_found)
        if datastore_info is not None and datastore_info.fields_dict is not None:
            if not return_list:
                return datastore_info.fields_dict
            else:
                return [field_info.to_ckan_dict() for field_info in datastore_info.fields_dict.values()]
        else:
            return None

    def get_resource_view_list_or_request(self, resource_id:str, error_not_found:bool=True) -> Union[List[CkanViewInfo],None]:
        """
        Returns either the resource view list which was already received or emits a new query for this information.

        :param resource_id:
        :param error_not_found:
        :return:
        """
        resource_info = self.get_resource_info_or_request_of_id(resource_id, error_not_found=error_not_found)
        if resource_info is None:
            return None
        elif not resource_info.view_is_full_list:
            resource_info.update_view(self.resource_view_list(resource_id))
        return list(resource_info.views.values())

    def get_package_info_or_request(self, package_name:str, *,
                                    request_missing:bool=True, error_not_mapped:bool=False, error_not_found:bool=True,
                                    datastore_info:bool=None, resource_view_list:bool=None, organization_info:bool=None,
                                    license_list:bool=None,) -> Union[CkanPackageInfo,None]:
        """
        Get information on a Package if present in the map or perform request.

        :param package_name: package name or id
        :param request_missing: confirm to perform the request if the information is missing
        :param error_not_mapped: raise error if the resource is not mapped
        :return:
        """
        package_info = self.map.get_package_info(package_name, error_not_mapped=error_not_mapped)
        if package_info is not None:
            return package_info
        elif request_missing:
            self.map_resources(package_name, error_not_found=error_not_found,
                               datastore_info=datastore_info, resource_view_list=resource_view_list,
                               organization_info=organization_info, license_list=license_list)  # request DataStore information if parameterized for
            return self.map.get_package_info(package_name, error_not_mapped=error_not_mapped)
        else:
            return None

    def get_organization_info_or_request(self, organization_name:str, *,
                                         request_missing:bool=True, error_not_mapped:bool=False,
                                         error_not_found:bool=True) -> Union[CkanOrganizationInfo,None]:
        """
        Get information on a Package if present in the map or perform request.

        :param organization_name: organization name or id
        :param request_missing: confirm to perform the request if the information is missing
        :param error_not_mapped: raise error if the resource is not mapped
        :return:
        """
        organization_info = self.map.get_organization_info(organization_name, error_not_mapped=error_not_mapped)
        if organization_info is not None:
            return organization_info
        elif request_missing:
            try:
                return self.organization_show(organization_name)
            except CkanNotFoundError as e:
                if error_not_found:
                    raise e from e
                else:
                    return None
        else:
            return None

    ## API calls needed to make the map and auxiliary API functions  ------------------
    def _api_package_search(self, *, params:dict=None, owner_org:str=None, filter:dict=None, q:str=None,
                            include_private:bool=True, include_drafts:bool=False, sort:str=None,
                            facet:bool=None, limit:int=None, offset:int=None) -> List[CkanPackageInfo]:
        """
        API call to package_search.

        :param owner_org: ability to filter packages by owner_org
        :param filter: dict of filters to apply, which translate to the API fq argument
            fq documentation: any filter queries to apply. Note: +site_id:{ckan_site_id} is added to this string prior to the query being executed.
        :param q: the solr query. Optional. Default is '*:*'
        :param include_private: if True, private datasets will be included in the results. Only private datasets from the user’s organizations will be returned and sysadmins will be returned all private datasets. Optional, the default is False in the API
        :param include_drafts:  if True, draft datasets will be included in the results. A user will only be returned their own draft datasets, and a sysadmin will be returned all draft datasets. Optional, the default is False.
        :param sort: sorting of the search results. Optional. Default: 'score desc, metadata_modified desc'. As per the solr documentation, this is a comma-separated string of field names and sort-orderings.
        :param facet:  whether to enable faceted results. Default: True in API.
        :param limit: maximum number of results to return. Translatees to the API rows argument.
        :param offset: the offset in the complete result for where the set of returned datasets should begin. Translatees to the API start argument.
        :param params: other parameters to pass to package_search
        :return:
        """
        if params is None: params = {}
        if limit is None: limit = self.params.default_limit_list
        if limit is not None:
            params["rows"] = limit
        if offset is not None:
            params["start"] = offset
        if owner_org is None and use_ckan_owner_org_as_default:
            owner_org = self.owner_org
        if owner_org is not None:
            owner_org_info = self.get_organization_info_or_request(owner_org)
            owner_org = owner_org_info.id
            if filter is None: filter = {}
            filter["owner_org"] = owner_org
        if q is not None:
            params["q"] = q
        if filter is not None:
            params["fq"] = '+'.join([f"{key}:{value}" for key, value in filter.items()])
        if sort is not None:
            params["sort"] = sort
        if facet is not None:
            params["facet"] = facet  # what are facets? default was False in previous versions
        if include_private is not None:
            params["include_private"] = include_private
        if include_drafts is not None:
            params["include_drafts"] = include_drafts
        response = self._api_action_request("package_search", method=RequestType.Get, params=params)
        if response.dry_run:
            return []
        elif response.success:
            package_info_list = [CkanPackageInfo(e) for e in response.result["results"]]
            self.map._update_package_info(package_info_list)
            return package_info_list
        else:
            raise response.default_error(self)

    def _api_package_search_all(self, *, params:dict=None, owner_org:str=None, filter:dict=None, q:str=None,
                                include_private:bool=True, include_drafts:bool=False, sort:str=None,
                                facet:bool=None, limit:int=None, offset:int=None, search_all:bool=True) -> List[CkanPackageInfo]:
        """
        API call to package_search until an empty list is received.

        :see: _api_package_search()
        :param owner_org: ability to filter packages by owner_org
        :param filter: dict of filters to apply, which translate to the API fq argument
            fq documentation: any filter queries to apply. Note: +site_id:{ckan_site_id} is added to this string prior to the query being executed.
        :param q: the solr query. Optional. Default is '*:*'
        :param include_private: if True, private datasets will be included in the results. Only private datasets from the user’s organizations will be returned and sysadmins will be returned all private datasets. Optional, the default is False in the API
        :param include_drafts:  if True, draft datasets will be included in the results. A user will only be returned their own draft datasets, and a sysadmin will be returned all draft datasets. Optional, the default is False.
        :param sort: sorting of the search results. Optional. Default: 'score desc, metadata_modified desc'. As per the solr documentation, this is a comma-separated string of field names and sort-orderings.
        :param facet:  whether to enable faceted results. Default: True in API.
        :param limit: maximum number of results to return. Translatees to the API rows argument.
        :param offset: the offset in the complete result for where the set of returned datasets should begin. Translatees to the API start argument.
        :param params: other parameters to pass to package_search
        :return:
        """
        if params is None: params = {}
        responses = self._request_all_results_list(self._api_package_search, params=params, limit=limit, offset=offset,
                                                   owner_org=owner_org, filter=filter, q=q, sort=sort, facet=facet,
                                                   include_private=include_private, include_drafts=include_drafts,
                                                   search_all=search_all)
        return sum(responses, [])

    def package_search_all(self, *, params:dict=None, owner_org:str=None, filter:dict=None, q:str=None,
                                include_private:bool=True, include_drafts:bool=False, sort:str=None,
                                facet:bool=None, limit:int=None, offset:int=None, search_all:bool=True) -> List[CkanPackageInfo]:
        # function alias
        return self._api_package_search_all(params=params, owner_org=owner_org, filter=filter, q=q,
                                            include_private=include_private, include_drafts=include_drafts, sort=sort,
                                            facet=facet, limit=limit, offset=offset, search_all=search_all)


    def _api_package_show(self, package_id, *, params:dict=None) -> CkanPackageInfo:
        """
        API call to package_show. Returns the information on the package and the resources contained in the package.
        Not recommended for outer use because this method does not return information about the DataStores. Prefer the map_resources method.

        :see: map_resources()
        :param package_id: package id.
        :param params: See API documentation.
        :return:
        """
        if params is None: params = {}
        params["id"] = package_id
        response = self._api_action_request("package_show", method=RequestType.Get, params=params)
        if response.success:
            package_info = CkanPackageInfo(response.result)
            # update map
            self.map._update_package_info(package_info)
            return package_info.copy()
        elif response.status_code == 404 and response.success_json_loads and response.error_message["__type"] == "Not Found Error":
            raise CkanNotFoundError(self, "Package", response)
        else:
            raise response.default_error(self)

    def package_show(self, package_id, *, params:dict=None) -> CkanPackageInfo:
        # function alias
        return self._api_package_show(package_id=package_id, params=params)

    def _api_resource_show(self, resource_id, *, params:dict=None) -> CkanResourceInfo:
        """
        API call to resource_show. Returns the metadata on a resource.

        :param resource_id: resource id.
        :param params: See API documentation.
        :return:
        """
        if params is None: params = {}
        params["id"] = resource_id
        response = self._api_action_request("resource_show", method=RequestType.Get, params=params)
        if response.success:
            return CkanResourceInfo(response.result)
        elif response.status_code == 404 and response.success_json_loads and response.error_message["__type"] == "Not Found Error":
            raise CkanNotFoundError(self, "Resource", response)
        else:
            raise response.default_error(self)

    def resource_show(self, resource_id, *, params:dict=None) -> CkanResourceInfo:
        # function alias
        return self._api_resource_show(resource_id=resource_id, params=params)

    def _api_datastore_info(self, resource_id:str, *, params:dict=None, display_request_not_found:bool=True) -> CkanDataStoreInfo:
        """
        API call to datastore_info. Returns the information on the DataStore. Used to know the number of rows in a DataStore.

        :param resource_id: resource id.
        :param params: N/A
        :param display_request_not_found: whether to display the request in the command window, in case of a CkanNotFoundError.
            This option is recommended if you are testing whether the resource has a DataStore or not.
        :return:
        """
        if params is None: params = {}
        params["id"] = resource_id
        response = self._api_action_request("datastore_info", method=RequestType.Post, json=params)
        if response.success:
            datastore_info = CkanDataStoreInfo(response.result)
            self.map._update_datastore_info(datastore_info)
            return datastore_info.copy()
        elif response.status_code == 404 and response.success_json_loads and response.error_message["__type"] == "Not Found Error":
            raise CkanNotFoundError(self, "DataStore", response, display_request=display_request_not_found)
        else:
            raise response.default_error(self)

    def datastore_info(self, resource_id:str, *, params:dict=None, display_request_not_found:bool=True) -> CkanDataStoreInfo:
        # function alias
        return self._api_datastore_info(resource_id=resource_id, params=params, display_request_not_found=display_request_not_found)

    def _api_resource_view_list(self, resource_id:str, *, params:dict=None) -> List[CkanViewInfo]:
        """
        API call to resource_view_list.

        :param params: typically, the request can be limited to an organization with the owner_org parameter
        :return:
        """
        if params is None:
            params = {}
        params["id"] = resource_id
        response = self._api_action_request("resource_view_list", method=RequestType.Get, params=params)
        if response.success:
            view_info_list = [CkanViewInfo(view_dict) for view_dict in response.result]
            self.map._update_view_info(view_info_list, view_list=True)
            return copy.deepcopy(view_info_list)
        else:
            raise response.default_error(self)

    def resource_view_list(self, resource_id:str, *, params:dict=None) -> List[CkanViewInfo]:
        # function alias
        return self._api_resource_view_list(resource_id=resource_id, params=params)

    def _api_organization_show(self, id:str, *, params:dict=None) -> CkanOrganizationInfo:
        """
        API call to organization_show.

        :param id: organization id or name.
        :param params: typically, the request can be limited to an organization with the owner_org parameter
        :return:
        """
        if params is None: params = {}
        if id is not None:
            params["id"] = id
        response = self._api_action_request("organization_show", method=RequestType.Get, params=params)
        if response.success:
            organization_info = CkanOrganizationInfo(response.result)
            # update map
            self.map._update_organization_info(organization_info)
            return organization_info.copy()
        elif response.status_code == 404 and response.success_json_loads and response.error_message["__type"] == "Not Found Error":
            raise CkanNotFoundError(self, "Organization", response)
        else:
            raise response.default_error(self)

    def organization_show(self, id:str, *, params:dict=None) -> CkanOrganizationInfo:
        # function alias
        return self._api_organization_show(id=id, params=params)

    def _api_organization_list(self, *, params:dict=None, all_fields:bool=True,
                               include_users:bool=False,
                               limit:int=None, offset:int=None) -> Union[List[CkanOrganizationInfo], List[str]]:
        """
        API call to organization_list.

        :param params: typically, the request can be limited to an organization with the owner_org parameter
        :param all_fields: whether to return full information or only the organization names in a list
        :return:
        """
        if params is None: params = {}
        if limit is None: limit = self.params.default_limit_list
        if limit is not None:
            params["limit"] = limit
        if offset is not None:
            params["offset"] = offset
        params["all_fields"] = all_fields
        params["include_users"] = include_users
        response = self._api_action_request("organization_list", method=RequestType.Get, params=params)
        if response.success:
            if all_fields:
                # returns a list of dicts
                organization_list = [CkanOrganizationInfo(e) for e in response.result]
                # update map
                self.map._update_organization_info(organization_list)
                return copy.deepcopy(organization_list)
            else:
                # returns a list of organization names
                return response.result
        else:
            raise response.default_error(self)

    def _api_organization_list_all(self, *, params:dict=None, all_fields:bool=True,
                                   include_users:bool=False,
                                   limit:int=None, offset:int=None) -> Union[List[CkanOrganizationInfo], List[str]]:
        """
        API call to organization_list until an empty list is received.

        :see: _api_organization_list()
        :param params:
        :return:
        """
        if params is None: params = {}
        responses = self._request_all_results_list(self._api_organization_list, params=params, limit=limit, offset=offset,
                                                   all_fields=all_fields, include_users=include_users)
        self.map.organizations_listed_all = True
        self.map.organizations_listed_all_users = include_users
        return sum(responses, [])

    def organization_list_all(self, *, cancel_if_present:bool=False, params:dict=None,
                              all_fields:bool=True, include_users:bool=False,
                              limit:int=None, offset:int=None) -> Union[List[CkanOrganizationInfo], List[str]]:
        """
        API call to license_list.
        The call can be canceled if the list is already present (not recommended, rather use get_organization_info_or_request).

        :param params:
        :param cancel_if_present: option to cancel when list is already present.
        :return:
        """
        if self.map.organizations_listed_all and cancel_if_present \
                and self.map.organizations_listed_all_users == include_users:
            return list(self.map.organizations.values())
        else:
            return self._api_organization_list_all(params=params, all_fields=all_fields, include_users=include_users, limit=limit, offset=offset)

    def _api_license_list(self, *, params:dict=None) -> List[CkanLicenseInfo]:
        """
        API call to license_list.

        :param params:
        :return:
        """
        if params is None: params = {}
        response = self._api_action_request(f"license_list", method=RequestType.Post, json=params)
        if response.success:
            license_list = [CkanLicenseInfo(license_dict) for license_dict in response.result]
            # update map:
            self.map._update_license_info(license_list)
            return copy.deepcopy(license_list)
        else:
            raise response.default_error(self)

    def license_list(self, *, cancel_if_present:bool=True, params:dict=None) -> List[CkanLicenseInfo]:
        """
        API call to license_list. The call can be canceled if the list is already present.

        :param params:
        :param cancel_if_present: option to cancel when list is already present.
        :return:
        """
        if len(self.map.licenses) > 0 and cancel_if_present:
            return list(self.map.licenses.values())
        else:
            return self._api_license_list(params=params)

    def resource_is_datastore(self, resource_id:str) -> bool:
        """
        Basic test to know whether a resource is DataStore.

        :param resource_id:
        :return:
        """
        try:
            datastore_info = self.datastore_info(resource_id, display_request_not_found=False)
        except CkanNotFoundError as e:
            return False
        return True

    def get_package_page_url(self, package_name:str, *, error_not_found:bool=True, default_url:bool=False) -> str:
        """
        Get URL of package presentation page in CKAN (landing page).

        :param package_name:
        :param error_not_found:
        :param default_url: return url based on package name, even if it was not found.
        :return:
        """
        self._error_empty_url()
        initial_url = url_join(self.url, "dataset" + urlsep + package_name)
        if default_url and error_not_found:
            return initial_url
        package_info = self.get_package_info_or_request(package_name, error_not_found=error_not_found)
        if package_info is not None:
            url = url_join(self.url, "dataset" + urlsep + package_info.name)
        elif default_url:
            msg = f"Package {package_name} not found. Returning default URL {initial_url}."
            warn(msg)
            return initial_url
        else:
            url = None
        return url

    def get_resource_page_url(self, resource_name:str, package_name:str=None,
                              *, error_not_mapped:bool=True) -> str:
        """
        Get URL of resource presentation page in CKAN (landing page).

        :param package_name:
        :return:
        """
        self._error_empty_url()
        resource_info = self.map.get_resource_info(resource_name, package_name=package_name, error_not_mapped=error_not_mapped)
        if resource_info is not None:
            package_info = self.map.get_package_info(resource_info.package_id)
            url = url_join(self.url, "dataset" + urlsep + package_info.name + urlsep + "resource" + urlsep + resource_info.id)
        else:
            url = None
        return url

    def test_ckan_connection(self, raise_error:bool=False) -> bool:
        """
        Test if the CKAN URL aims to a CKAN server by testing the package_search API.
        This does not check authentication.
        """
        try:
            self.package_search_all(limit=1, search_all=False)
        except CkanActionError as e:
            if e.status_code == 220:
                if raise_error:
                    raise e from e
                else:
                    return False
            else:
                raise e from e
        return True

    def _api_user_show(self, *, params:dict=None) -> Union[CkanUserInfo,None]:
        """
        API call to user_show. With no params, returns the name of the current user logged in.

        :return: dict with information on the current user
        """
        if params is None: params = {}
        response = self._api_action_request("user_show", method=RequestType.Get, params=params, timeout=5)
        if response.success:
            user_info = CkanUserInfo(response.result)
            self.map._update_user_info(user_info)
            return user_info.copy()
        elif response.status_code == 404 and response.success_json_loads and response.error_message["__type"] == "Not Found Error":
            raise CkanNotFoundError(self, "User", response)
        else:
            raise response.default_error(self)

    def query_current_user(self, *, verbose:bool=None, error_not_found:bool=False) -> Union[CkanUserInfo,None]:
        if verbose is None:
            verbose = self.params.verbose_extra
        try:
            user_info = self._api_user_show()
        except CkanNotFoundError as e:
            if error_not_found:
                raise e from e
            else:
                user_info = None
        if verbose:
            if user_info is not None:
                print("Authenticated as " + user_info.name)
            else:
                print("User not authenticated")
        return user_info

    def test_ckan_login(self, *, raise_error:bool=False, verbose:bool=None,
                        empty_key_connected:bool=False) -> bool:
        """
        Test if your login leads to a user account.

        :param raise_error: option to raise an error if no account was detected
        :param verbose: option to display username in console
        :param empty_key_connected: option to ignore the test if the API key is empty
        """
        user_info = self.query_current_user(verbose=verbose, error_not_found=raise_error and not empty_key_connected)
        if user_info is None:
            if self.apikey.is_empty():
                if raise_error and not empty_key_connected:
                    raise ConnectionError("No API key loaded")
                return empty_key_connected
            else:
                if raise_error:
                    raise ConnectionError("The current API key did not authenticate a user")
                return False
        else:
            return True


    ## List users and groups
    def _api_user_list(self, *, q:str=None, email:str=None,
                       params:dict=None) -> List[CkanUserInfo]:
        """
        API call to user_list.

        :param params:
        :return:
        """
        if params is None: params = {}
        if q is not None:
            params["q"] = q
        if email is not None:
            params["email"] = email
        response = self._api_action_request(f"user_list", method=RequestType.Post, json=params)
        if response.success:
            user_list = [CkanUserInfo(user_dict) for user_dict in response.result]
            # update map:
            self.map._update_user_info(user_list)
            return copy.deepcopy(user_list)
        else:
            raise response.default_error(self)

    def user_list(self, *, cancel_if_present:bool=False, q:str=None, email:str=None,
                  params:dict=None) -> List[CkanUserInfo]:
        """
        API call to user_list. The call can be canceled if the list is already present.

        :param params:
        :param cancel_if_present: option to cancel when list is already present.
        :return:
        """
        if q is None and email is None and params is None:
            if self.map.users_listed_all > 0 and cancel_if_present:
                return list(self.map.users.values())
            else:
                result = self._api_user_list(params=params)
                self.map.users_listed_all = True
                return result
        else:
            return self._api_user_list(q=q, email=email, params=params)

    def _api_package_collaborator_list(self, package_id:str, *, params:dict=None,
                                       cancel_if_present:bool=False) -> Dict[str,CkanCollaboration]:
        """
        API call to package_collaborator_list.

        :param params:
        :return:
        """
        if cancel_if_present:
            package_info = self.get_package_info_or_request(package_id)
            if package_info.collaborators is not None:
                return package_info.collaborators
        if params is None: params = {}
        params["id"] = package_id
        response = self._api_action_request(f"package_collaborator_list", method=RequestType.Post, json=params)
        if response.success:
            package_info = self.get_package_info_or_request(package_id)
            package_info.collaborators = {}
            for collaborator_dict in response.result:
                assert (collaborator_dict["package_id"] == package_id)
                package_info.collaborators[collaborator_dict["user_id"]] = CkanCollaboration(d=collaborator_dict)
            return package_info.collaborators
        else:
            raise response.default_error(self)

    def package_collaborator_list(self, package_id:str, *, params:dict=None,
                                  cancel_if_present:bool=False) -> Dict[str,CkanCollaboration]:
        return self._api_package_collaborator_list(package_id=package_id, params=params,
                                                   cancel_if_present=cancel_if_present)

    def _api_group_list(self, *, limit:int=None, offset:int=0, groups:List[str]=None,
                        all_fields:bool=True, include_users:bool=True,
                        params:dict=None) -> Union[List[CkanGroupInfo], List[str]]:
        """
        API call to group_list.

        :param params:
        :return:
        """
        if params is None: params = {}
        if limit is None:
            limit = self.params.default_limit_list
        if limit is not None:
            params["limit"] = limit
        if offset is not None:
            params["offset"] = offset
        if groups is not None:
            params["groups"] = groups
        params["include_users"] = include_users
        all_fields = all_fields or include_users
        params["all_fields"] = all_fields
        response = self._api_action_request(f"group_list", method=RequestType.Post, json=params)
        if response.success:
            if all_fields:
                group_list = [CkanGroupInfo(group_dict) for group_dict in response.result]
                # update map:
                if include_users:
                    for group_info in group_list:
                        user_list = [CkanUserInfo(user_dict) for user_dict in group_info.details["users"]]
                        self.map._update_user_info(user_list)
                        group_info.user_members = {user_info.id: CkanCapacity.from_str(user_dict["capacity"]) for user_info, user_dict in zip(user_list, group_info.details["users"])}
                self.map._update_group_info(group_list)
                return copy.deepcopy(group_list)
            else:
                return response.result  # list of names
        else:
            raise response.default_error(self)

    def group_list(self, *, limit:int=None, offset:int=0, groups:List[str]=None,
                        all_fields:bool=True, include_users:bool=True,
                        params:dict=None) -> List[CkanGroupInfo]:
        # function alias
        return self._api_group_list(groups=groups, all_fields=all_fields, include_users=include_users,
                                    limit=limit, offset=offset, params=params)

    def _api_group_list_all(self, *, all_fields:bool=True, include_users:bool=True, params:dict=None,
                            limit:int=None, offset:int=None) -> Union[List[CkanUserInfo], List[str]]:
        """
        API call to group_list until an empty list is received.

        :see: _api_group_list()
        :param params:
        :return:
        """
        if params is None: params = {}
        responses = self._request_all_results_list(self._api_group_list, params=params, limit=limit, offset=offset,
                                                   all_fields=all_fields, include_users=include_users)
        self.map.groups_listed_all = True
        return sum(responses, [])

    def group_list_all(self, *, all_fields:bool=True, include_users:bool=True,
                       cancel_if_present:bool=False, params:dict=None,
                       limit:int=None, offset:int=None) -> Union[List[CkanGroupInfo], List[str]]:
        """
        API call to group_list.
        The call can be canceled if the list is already present (not recommended, rather use get_organization_info_or_request).

        :param params:
        :param cancel_if_present: option to cancel when list is already present.
        :return:
        """
        if self.map.groups_listed_all and cancel_if_present:
            return list(self.map.groups.values())
        else:
            return self._api_group_list_all(params=params, all_fields=all_fields, include_users=include_users, limit=limit, offset=offset)

    def map_user_rights(self, cancel_if_present:bool=True):
        """
        Map user and group access rights to the packages currently mapped by CKAN
        :return:
        """
        current_user = self.query_current_user()
        self.group_list_all(cancel_if_present=cancel_if_present)
        self.user_list(cancel_if_present=cancel_if_present)
        for package_id, package_info in self.map.packages.items():
            if current_user is not None:
                self.package_collaborator_list(package_id, cancel_if_present=cancel_if_present)
            # merge collaborators with groups of the package
            if package_info.collaborators is not None:
                package_info.user_access = package_info.collaborators.copy()
                for group in package_info.groups:
                    group_info = self.map.groups[group.id]
                    for user_id, user_capacity in group_info.user_members.items():
                        if user_id not in package_info.user_access.keys():
                            package_info.user_access[user_id] = CkanCollaboration(user_capacity, None, group_id=group.id)
        return self.map



#!python3
# -*- coding: utf-8 -*-
"""
Data model to represent a CKAN database architecture
"""

ckan_tags_sep = ','   # separator for tags, fields in primary key, indexes, queries
ckan_timestamp_sep = 'T'  # separator for timestamp ISO format
environ_keyword = "environ"




#!python3
# -*- coding: utf-8 -*-
"""

"""
from typing import List, Dict, Union
import argparse

from ckanapi_harvesters.auxiliary.proxy_config import ProxyConfig
from ckanapi_harvesters.auxiliary.ckan_map import CkanMap
from ckanapi_harvesters.auxiliary import ckan_configuration
from ckanapi_harvesters.policies.data_format_policy_errors import DataPolicyError
from ckanapi_harvesters.policies.data_format_policy import CkanPackageDataFormatPolicy

from ckanapi_harvesters.auxiliary.ckan_api_key import CkanApiKey
from ckanapi_harvesters.ckan_api.ckan_api_2_readonly import CkanApiReadOnlyParams
from ckanapi_harvesters.ckan_api.ckan_api_2_readonly import CkanApiReadOnly

ckan_default_policy_keyword = "default"


class CkanApiPolicyParams(CkanApiReadOnlyParams):
    def __init__(self, *, proxies:Union[str,dict,ProxyConfig]=None,
                 ckan_headers:dict=None, http_headers:dict=None):
        super().__init__(proxies=proxies, ckan_headers=ckan_headers, http_headers=http_headers)
        self.policy_check_pre: bool = False
        self.policy_check_post: bool = True
        self.verbose_policy: bool = True

    def copy(self, new_identifier:str=None, *, dest=None):
        if dest is None:
            dest = CkanApiPolicyParams()
        super().copy(dest=dest)
        dest.policy_check_pre = self.policy_check_pre
        dest.policy_check_post = self.policy_check_post
        dest.verbose_policy = self.verbose_policy
        return dest


class CkanApiPolicy(CkanApiReadOnly):
    def __init__(self, url: str = None, *, proxies:Union[str,dict,ProxyConfig] = None,
                 apikey: Union[str,CkanApiKey] = None, apikey_file: str = None,
                 owner_org: str = None, params:CkanApiPolicyParams=None,
                 map:CkanMap=None, policy: CkanPackageDataFormatPolicy = None, policy_file: str = None,
                 identifier=None):
        """
        CKAN Database API interface to CKAN server with helper functions using pandas DataFrames.

        :param url: url of the CKAN server
        :param proxies: proxies to use for requests
        :param apikey: way to provide the API key directly (optional)
        :param apikey_file: path to a file containing a valid API key in the first line of text (optional)
        :param policy: data format policy to use with policy_check function
        :param policy_file: path to a JSON file containing the data format policy to use with policy_check function
        :param owner_org: name of the organization to limit package_search (optional)
        :param params: other connection/behavior parameters
        :param map: map of known resources
        :param policy: data format policy to be used with the policy_check function.
        :param policy_file: path to a JSON file containing the data format policy to load.
        :param identifier: identifier of the ckan client
        """
        super().__init__(url=url, proxies=proxies, apikey_file=apikey_file, apikey=apikey,
                         owner_org=owner_org, map=map, identifier=identifier)
        self.policy: Union[CkanPackageDataFormatPolicy,None] = policy
        self.policy_source: Union[str,None] = None
        if policy_file is not None:
            self.load_policy(policy_file, base_dir=None)
        self.default_policy_load_on_map: bool = False
        if params is None:
            params = CkanApiPolicyParams()
        if proxies is not None:
            params.proxies = proxies
        self.params: CkanApiPolicyParams = params

    def copy(self, new_identifier: str = None, *, dest=None):
        if dest is None:
            dest = CkanApiPolicy()
        super().copy(new_identifier=new_identifier, dest=dest)
        if self.policy is not None:
            dest.policy = self.policy.copy()
        dest.default_policy_load_on_map = self.default_policy_load_on_map
        return dest

    def set_verbosity(self, verbosity:bool=True, verbose_extra:bool=None) -> None:
        """
        Enable/disable full verbose output

        :param verbosity: boolean. Cannot be None
        :return:
        """
        super().set_verbosity(verbosity=verbosity, verbose_extra=verbose_extra)
        if verbose_extra is not None:
            self.params.verbose_policy = verbose_extra

    def _setup_cli_ckan_parser(self, parser:argparse.ArgumentParser=None) -> argparse.ArgumentParser:
        # overload adding support to load a policy from a file
        parser = super()._setup_cli_ckan_parser(parser=parser)
        parser.add_argument("--policy-file", type=str,
                            help="Path to a file containing the CKAN data format policy (json format)")
        return parser

    def _cli_ckan_args_apply(self, args: argparse.Namespace, *, base_dir:str=None,
                             error_not_found:bool=True, default_proxies:dict=None, proxy_headers:dict=None,
                             proxies:dict=None, headers:dict=None) -> None:
        # overload adding support to load a policy from a file
        super()._cli_ckan_args_apply(args=args, base_dir=base_dir, error_not_found=error_not_found,
                                     default_proxies=default_proxies, proxy_headers=proxy_headers)
        if proxies is None:
            proxies = self.params.proxies
        if args.policy_file is not None:
            self.load_policy(args.policy_file, proxies=proxies, headers=headers, error_not_found=error_not_found)

    def query_default_policy(self, *, error_not_found:bool=False, load_error:bool=True) -> Union[CkanPackageDataFormatPolicy,None]:
        """
        Download default policy and return it without loading it in the policy attribute.

        :param error_not_found:
        :return:
        """
        self.map_resources(ckan_configuration.configuration_package_name, error_not_found=error_not_found, load_policy=False)  # load_policy=False or else infinite loop
        resource_info = self.map.get_resource_info(ckan_configuration.policy_resource, ckan_configuration.configuration_package_name,
                                                   error_not_mapped=error_not_found)
        if resource_info is not None:
            resource_id = resource_info.id
            url = resource_info.download_url
            _, response = self.resource_download(resource_id)
            payload = response.text
            return CkanPackageDataFormatPolicy.from_jsons(payload, source_file=url, load_error=load_error)
        else:
            return None

    def load_default_policy(self, *,
                            error_not_found:bool=False, load_error:bool=True, cancel_if_present:bool=False,
                            force:bool=False) -> Union[CkanPackageDataFormatPolicy,None]:
        """
        Function to load the default data format policy from the CKAN server.
        The default policy is defined in ckan_configuration

        :param error_not_found:
        :param cancel_if_present:
        :param force:
        :return:
        """
        if force:
            self.policy = None
        if self.policy_source == ckan_default_policy_keyword and cancel_if_present:
            return self.policy
        self.policy = self.query_default_policy(error_not_found=error_not_found, load_error=load_error)
        self.policy_source = ckan_default_policy_keyword
        return self.policy

    def load_policy(self, policy_file: str, base_dir: str = None, proxies:dict=None, headers:dict=None,
                    error_not_found: bool = True) -> CkanPackageDataFormatPolicy:
        """
        Load the CKAN data format policy from file (JSON format).

        :param policy_file: path to the policy file
        :param base_dir: base directory, if the apikey_file is a relative path
        :return:
        """
        if proxies is None:
            proxies = self.params.proxies
        if policy_file is None:
            policy_file = ckan_default_policy_keyword  # set to "default"
        if policy_file is not None and policy_file.lower() == ckan_default_policy_keyword:  # if equals "default"
            return self.load_default_policy(error_not_found=error_not_found, force=True, cancel_if_present=False)
        self.policy = CkanPackageDataFormatPolicy.from_json(policy_file, base_dir=base_dir, proxies=proxies, headers=headers,
                                                            error_not_found=error_not_found)
        self.policy_source = policy_file
        return self.policy

    def policy_check(self, package_list:Union[str,List[str]]=None, policy: CkanPackageDataFormatPolicy=None,
                     *, buffer:Dict[str, List[DataPolicyError]]=None, raise_error:bool=False,
                     verbose:bool=None, auto_update:bool=True) -> bool:
        """
        Enforce policy on mapped packages

        :param policy:
        :return:
        """
        success = True
        if package_list is None:
            package_list = self.map.packages.keys()  # check on all packages
        elif isinstance(package_list, str):
            package_list = [package_list]
        if policy is None:
            policy = self.policy
        if verbose is None:
            verbose = self.params.verbose_policy
        if policy is None:
            # no policy loaded at all
            return True
        if verbose:
            print(f"Testing policy {policy.label}")
        for package_name in package_list:
            package_info = self.get_package_info_or_request(package_name)
            package_buffer: List[DataPolicyError] = []
            success &= policy.policy_check_package(package_info, display_message=verbose,
                                                   package_buffer=package_buffer, raise_error=raise_error)
            if auto_update:
                policy.package_update_scores(self, package_info, package_buffer, raise_error=raise_error)
            if buffer is not None:
                buffer[package_info.name] = package_buffer
        if verbose:
            print(f"Data format policy {policy.label} success: {success}")
        return success

    def set_default_map_mode(self, datastore_info:bool=None, resource_view_list:bool=None,
                             organization_info:bool=None, license_list:bool=None,
                             load_policy:bool=None) -> None:
        super().set_default_map_mode(datastore_info=datastore_info, resource_view_list=resource_view_list,
                                     organization_info=organization_info, license_list=license_list)
        if load_policy is None:
            load_policy = self.default_policy_load_on_map
        self.default_policy_load_on_map = load_policy

    def map_resources(self, package_list:Union[str, List[str]]=None, *, params:dict=None,
                      datastore_info:bool=None, resource_view_list:bool=None, organization_info:bool=None, license_list:bool=None,
                      only_missing:bool=True, error_not_found:bool=True,
                      owner_org:str=None, load_policy:bool=None) -> CkanMap:
        # overload including a call to load the default data format policy
        self.set_default_map_mode(load_policy=load_policy)
        map = super().map_resources(package_list=package_list, params=params, datastore_info=datastore_info,
                              resource_view_list=resource_view_list, organization_info=organization_info,
                              license_list=license_list, only_missing=only_missing, error_not_found=error_not_found,
                              owner_org=owner_org)
        load_policy = self.default_policy_load_on_map
        if load_policy:
            self.load_default_policy(cancel_if_present=True, load_error=False)
        return map




#!python3
# -*- coding: utf-8 -*-
"""
Harvest from a PostgreSQL database using sqlalchemy
"""
from typing import Union, List, Any, Dict, Tuple
from types import SimpleNamespace
from collections import OrderedDict
from warnings import warn
import urllib.parse

import pandas as pd


from ckanapi_harvesters.auxiliary.lazy_imports import sqlalchemy, lazy_import_sqlalchemy
from ckanapi_harvesters.auxiliary.lazy_imports import psycopg2, lazy_import_psycopg2
from ckanapi_harvesters.harvesters.harvester_errors import (HarvesterRequirementError, HarvesterArgumentRequiredError)
from ckanapi_harvesters.harvesters.harvester_abc import TableHarvesterABC, DatasetHarvesterABC, DatabaseHarvesterABC
from ckanapi_harvesters.harvesters.harvester_model import FieldMetadata, TableMetadata, DatasetMetadata
from ckanapi_harvesters.harvesters.harvester_params import DatabaseParams
from ckanapi_harvesters.harvesters.postgre_params import DatasetParamsPostgreSchema, TableParamsPostgre
from ckanapi_harvesters.auxiliary.urls import url_join, url_insert_login
from ckanapi_harvesters.auxiliary.ckan_auxiliary import ssl_arguments_decompose, sql_varname_escape
from ckanapi_harvesters.auxiliary.ckan_auxiliary import parse_geometry_native_type
from ckanapi_harvesters.auxiliary.ckan_errors import UrlError
from ckanapi_harvesters.auxiliary.error_level_message import ContextErrorLevelMessage, ErrorLevel
from ckanapi_harvesters.harvesters.data_cleaner.data_cleaner_abc import CkanDataCleanerABC
from ckanapi_harvesters.harvesters.data_cleaner.data_cleaner_upload_2_geom import CkanDataCleanerUploadGeom

postgre_type_mapper = {}


class DatabaseHarvesterPostgre(DatabaseHarvesterABC):
    """
    This class manages the connection to a PostgreSQL database server.
    It can list schemas (corresponding to CKAN datasets).
    """
    def __init__(self, params:DatabaseParams=None):
        super().__init__(params)
        global sqlalchemy, psycopg2
        sqlalchemy = lazy_import_sqlalchemy()
        psycopg2 = lazy_import_psycopg2()
        if sqlalchemy.Engine is None:
            raise HarvesterRequirementError("sqlalchemy", "postgre")
        if psycopg2 is None:
            raise HarvesterRequirementError("psycopg2", "postgre")
        self.alchemy_engine: Union[sqlalchemy.Engine,None] = None
        self.alchemy_connection: Union[sqlalchemy.Connection,None] = None
        if self.params.auth_url is None and self.params.host is None and self.params.port is None and self.params.database is None:
            raise HarvesterArgumentRequiredError("auth-url", "postgre", "This argument defines the url used to authenticate.")

    @staticmethod
    def init_from_options_string(options_string:str, base_dir:str=None) -> "DatabaseHarvesterPostgre":
        params = DatabaseParams()
        params.parse_options_string(options_string, base_dir=base_dir)
        return DatabaseHarvesterPostgre(params)

    def copy(self, *, dest=None):
        if dest is None:
            dest = DatabaseHarvesterPostgre()
        return super().copy(dest=dest)

    def connect(self, *, cancel_if_connected:bool=True) -> Any:
        if cancel_if_connected and self.alchemy_engine is not None:
            return self.alchemy_engine
        else:
            if self.alchemy_engine is not None:
                self.alchemy_connection.close()
                self.alchemy_engine.dispose()
                self.alchemy_connection = None
                self.alchemy_engine = None
            ssl, ssl_certfile = ssl_arguments_decompose(self.params.verify_ca)
            auth_url = self.params.auth_url
            if auth_url is None:
                if self.params.url is not None:
                    auth_url = self.params.url
                elif self.params.host is not None:
                    auth_url = f"postgresql+psycopg2://{self.params.host}"
                    if self.params.port is not None:
                        auth_url += f":{self.params.port}"
                else:
                    raise UrlError("No Postgre URL provided")
                if self.params.auth_url_suffix is not None:
                    auth_url = url_join(auth_url, self.params.auth_url_suffix)
                elif self.params.database is not None:
                    auth_url = url_join(auth_url, self.params.database)
                self.params.auth_url = auth_url
            auth_url_with_login = url_insert_login(auth_url, self.params.login)
            self.alchemy_engine = sqlalchemy.create_engine(auth_url_with_login)
                                                           # ssl=ssl, tlscafile=ssl_certfile,
                                                           # timeoutMS=self.params.timeout*1000.0 if self.params.timeout is not None else None)
            self.alchemy_connection = self.alchemy_engine.connect()
            if self.params.host is None and self.params.port is None:
                # complete with host and port parsed by sqlalchemy
                parsed_url = urllib.parse.urlparse(auth_url)
                self.params.host, self.params.port = parsed_url.hostname, parsed_url.port
            return self.alchemy_engine

    def is_connected(self) -> bool:
        return self.alchemy_engine is not None

    def disconnect(self) -> None:
        if self.alchemy_engine is not None:
            if self.alchemy_connection is not None:
                self.alchemy_connection.close()
            try:
                self.alchemy_engine.dispose()
            except Exception as e:
                msg = f"Error while disconnecting PostgreSQL alchemy engine {self.params.url}: " + str(e)
                warn(msg)
            self.alchemy_engine = None
            self.alchemy_connection = None

    def check_connection(self, *, new_connection:bool=False, raise_error:bool=False) -> Union[None, ContextErrorLevelMessage]:
        try:
            self.connect(cancel_if_connected=not new_connection)
            remote_collections = self.list_datasets(return_metadata=False)
        except Exception as e:
            if raise_error:
                raise e from e
            else:
                return ContextErrorLevelMessage("Postgre Harvester", ErrorLevel.Error, f"Failed to connect to {self.params.auth_url}: {e}")

    def get_dataset_harvester(self, dataset_name: str) -> "DatasetHarvesterPostgre":
        params_dataset = self.params.copy(dest=DatasetParamsPostgreSchema())
        params_dataset.dataset = dataset_name
        dataset_harvester = DatasetHarvesterPostgre(params_dataset)
        self.copy(dest=dataset_harvester)
        dataset_harvester.params = params_dataset
        dataset_harvester._finalize_connection()
        return dataset_harvester

    def list_datasets(self, return_metadata: bool = True) -> Union[List[str], OrderedDict[str, DatasetMetadata]]:
        self.connect()
        query = "SELECT schema_name FROM information_schema.schemata;"
        df_schemas = pd.read_sql(query, self.alchemy_engine)
        dataset_list = df_schemas["schema_name"].tolist()
        if return_metadata:
            return OrderedDict([(name, self.get_dataset_harvester(name).query_dataset_metadata()) for name in dataset_list])
        else:
            return dataset_list


class DatasetHarvesterPostgre(DatabaseHarvesterPostgre, DatasetHarvesterABC):
    """
    A CKAN dataset corresponds to a PostgreSQL schema (set of tables).
    """
    def __init__(self, params:DatasetParamsPostgreSchema=None):
        super().__init__(params)
        self.dataset_metadata: Union[DatasetMetadata, None] = None  # DatasetHarvesterABC
        if self.params.dataset is None:
            raise HarvesterArgumentRequiredError("dataset", "postgre", "This argument defines the Postgre schema to be used")

    @staticmethod
    def init_from_options_string(options_string:str, base_dir:str=None) -> "DatasetHarvesterPostgre":
        params = DatasetParamsPostgreSchema()
        params.parse_options_string(options_string, base_dir=base_dir)
        return DatasetHarvesterPostgre(params)

    def _finalize_connection(self):
        if super().is_connected():
            pass

    def connect(self, *, cancel_if_connected:bool=True) -> Any:
        if not (cancel_if_connected and self.is_connected()):
            super().connect(cancel_if_connected=cancel_if_connected)
            self._finalize_connection()
        return self.alchemy_connection

    def is_connected(self) -> bool:
        return super().is_connected()

    def disconnect(self) -> None:
        if super().is_connected():
            super().disconnect()

    def check_connection(self, *, new_connection: bool = False, raise_error: bool = False) -> Union[None, ContextErrorLevelMessage]:
        try:
            super().check_connection(new_connection=new_connection, raise_error=raise_error)
            tables_list = self.list_tables(return_metadata=False)
        except Exception as e:
            if raise_error:
                raise e from e
            else:
                return ContextErrorLevelMessage("Postgre Harvester", ErrorLevel.Error,
                                                f"Failed to connect to {self.params.auth_url}: {e}")

    def query_dataset_metadata(self, cancel_if_present:bool=True) -> DatasetMetadata:
        self.connect()
        if cancel_if_present and self.dataset_metadata is not None:
            return self.dataset_metadata
        else:
            # query schema comment
            postgre_schema_name = self.params.dataset
            query = f"""
            SELECT 
                n.nspname AS {postgre_schema_name},
                d.description AS schema_comment
            FROM 
                pg_namespace n
            LEFT JOIN 
                pg_description d ON n.oid = d.objoid
            WHERE 
                n.nspname = '{postgre_schema_name}';
            """
            table_df = pd.read_sql(query, self.alchemy_engine)
            schema_comment = table_df.iloc[0]['schema_comment']
            self.dataset_metadata = DatasetMetadata()
            self.dataset_metadata.name = self.params.dataset
            self.dataset_metadata.description = schema_comment
            self.dataset_metadata.tables = self.list_tables(return_metadata=True)
            return self.dataset_metadata

    def clean_dataset_metadata(self) -> DatasetMetadata:
        return self.query_dataset_metadata().copy()

    def get_table_harvester(self, table_name:str) -> "TableHarvesterPostgre":
        params_table = self.params.copy(dest=TableParamsPostgre())
        if self.params.options_string is not None:
            # reparse options_string for table-specific arguments
            params_table.parse_options_string(self.params.options_string, base_dir=self.params.base_dir)
        params_table.table = table_name
        table_harvester = TableHarvesterPostgre(params_table)
        self.copy(dest=table_harvester)
        table_harvester.params = params_table
        table_harvester._finalize_connection()
        return table_harvester

    def list_tables(self, return_metadata:bool=True) -> Union[List[str], OrderedDict[str, TableMetadata]]:
        self.connect()
        postgre_schema_name = self.params.dataset
        query = f"""
        SELECT table_name
        FROM information_schema.tables
        WHERE table_schema = '{postgre_schema_name}'
          AND table_type = 'BASE TABLE';
        """
        df_tables = pd.read_sql(query, self.alchemy_engine)
        tables_list = df_tables["table_name"].tolist()
        if return_metadata:
            return OrderedDict([(name, self.get_table_harvester(name).query_table_metadata()) for name in tables_list])
        else:
            return tables_list


class TableHarvesterPostgre(DatasetHarvesterPostgre, TableHarvesterABC):
    """
    A CKAN table (DataStore) corresponds to a PostgreSQL table.
    """
    _default_upload_fun = None
    _default_primary_key = None

    def __init__(self, params:TableParamsPostgre=None):
        super().__init__(params)
        self.params: TableParamsPostgre = params
        self.table_metadata: Union[TableMetadata, None] = None  # TableHarvesterABC
        if self.params.file_url_attr is not None:
            # File/URL attribute has priority over CLI
            self.params.table = self.params.file_url_attr
        if self.params.table is None:
            raise HarvesterArgumentRequiredError("table", "postgre", "This argument defines the Postgre table used")

    @staticmethod
    def init_from_options_string(options_string:str, *, base_dir:str=None, file_url_attr:str=None) -> "TableHarvesterPostgre":
        params = TableParamsPostgre()
        params.parse_options_string(options_string, file_url_attr=file_url_attr, base_dir=base_dir)
        return TableHarvesterPostgre(params)

    def copy(self, *, dest=None):
        if dest is None:
            dest = TableHarvesterPostgre()
        super().copy(dest=dest)
        return dest

    def _finalize_connection(self):
        super()._finalize_connection()
        if super().is_connected():
            pass

    def connect(self, *, cancel_if_connected:bool=True) -> Any:
        if not (cancel_if_connected and self.is_connected()):
            super().connect()
            self._finalize_connection()
        return self.alchemy_engine

    def check_connection(self, *, new_connection:bool=False, raise_error:bool=False) -> Union[None, ContextErrorLevelMessage]:
        super().check_connection(new_connection=new_connection, raise_error=raise_error)

    def query_table_metadata(self, cancel_if_present:bool=True) -> TableMetadata:
        self.connect()
        if cancel_if_present and self.table_metadata is not None:
            return self.table_metadata
        else:
            postgre_schema_name = self.params.dataset
            postgre_table_name = self.params.table
            # request comment on table
            query = f"""
            SELECT 
                obj_description('{postgre_schema_name}.{postgre_table_name}'::regclass) AS table_comment;
            """
            table_df = pd.read_sql(query, self.alchemy_engine)
            table_comment = table_df.iloc[0]['table_comment']
            # request information on fields
            query = f"""
            SELECT
                cols.column_name,
                cols.ordinal_position,
                cols.is_nullable = 'NO' AS is_not_null,
                cols.data_type AS apparent_data_type,
                COALESCE(pt.typname, cols.data_type) AS full_data_type, -- Use user-defined type name if available
                pgd.description AS column_comment,
                EXISTS (
                    SELECT 1
                    FROM pg_index i
                    JOIN pg_attribute a ON a.attnum = ANY(i.indkey)
                    WHERE i.indrelid = ('{postgre_schema_name}.' || cols.table_name)::regclass
                      AND a.attname = cols.column_name
                ) AS is_indexed,
                con.constraint_type = 'UNIQUE' AS is_unique
            FROM
                information_schema.columns AS cols
            LEFT JOIN
                pg_catalog.pg_statio_all_tables AS st
                ON cols.table_schema = st.schemaname AND cols.table_name = st.relname
            LEFT JOIN
                pg_catalog.pg_description AS pgd
                ON pgd.objoid = st.relid AND pgd.objsubid = cols.ordinal_position
            LEFT JOIN
                information_schema.key_column_usage AS kcu
                ON cols.table_schema = kcu.table_schema
                   AND cols.table_name = kcu.table_name
                   AND cols.column_name = kcu.column_name
            LEFT JOIN
                information_schema.table_constraints AS con
                ON kcu.constraint_name = con.constraint_name
                   AND kcu.table_schema = con.table_schema
            LEFT JOIN
                pg_type pt
                ON cols.udt_name = pt.typname -- Match user-defined type name
            LEFT JOIN
                pg_namespace pn
                ON pt.typnamespace = pn.oid
            WHERE
                cols.table_schema = '{postgre_schema_name}'
                AND cols.table_name = '{postgre_table_name}'
            ORDER BY
                cols.ordinal_position;
            """
            # DataFrame with columns: ["column_name", "order", "is_not_null", "apparent_data_type", "full_data_type", "description", "is_indexed", "is_unique"]
            fields_df = pd.read_sql(query, self.alchemy_engine)
            fields_df.set_index("column_name", inplace=True, drop=False)
            assert(fields_df.index.is_unique)  # verify integrity
            # querying details on column types
            fields_df["definitive_data_type"] = fields_df["full_data_type"]
            fields_df["geo_type"] = ""
            fields_df["srid"] = 0
            # PostGIS geometry type
            if any(fields_df["full_data_type"] == "geometry"):
                query = f"""
                SELECT
                  f_geometry_column,
                  type,
                  srid
                FROM geometry_columns
                WHERE f_table_schema = '{postgre_schema_name}'
                  AND f_table_name = '{postgre_table_name}';
                """
                geo_df = pd.read_sql(query, self.alchemy_engine)
                for row_loc, row in geo_df.iterrows():
                    column_name = row["f_geometry_column"]
                    fields_df.loc[column_name, "definitive_data_type"] = f"geometry({row['type']}, {row['srid']})"
                    fields_df.loc[column_name, "geo_type"] = row['type']
                    fields_df.loc[column_name, "geo_srid"] = row['srid']
            # query primary key
            query = f"""
            SELECT
                a.attname AS column_name
            FROM
                pg_index i
            JOIN
                pg_attribute a ON a.attnum = ANY(i.indkey)
            WHERE
                i.indrelid = '{postgre_schema_name}.{postgre_table_name}'::regclass
                AND i.indisprimary;
            """
            primary_key_df = pd.read_sql(query, self.alchemy_engine)
            primary_key = primary_key_df["column_name"].tolist()
            if len(primary_key) == 0:
                primary_key = None
            self.table_metadata = TableMetadata()
            self.table_metadata.name = self.params.table
            self.table_metadata.description = table_comment
            self.table_metadata.fields = OrderedDict()
            for row_loc, row in fields_df.iterrows():
                field_metadata = FieldMetadata()
                field_metadata.name = row["column_name"]
                field_metadata.data_type = row["definitive_data_type"]
                field_metadata.harvester_attrs["datatype_keyword"] = row["full_data_type"]
                field_metadata.internal_attrs.geometry_as_source = row["geo_srid"] > 0
                field_metadata.internal_attrs.geometry_type = row["geo_type"] if row["geo_type"] else None
                field_metadata.internal_attrs.epsg_source = row["geo_srid"] if row["geo_srid"] > 0 else None
                field_metadata.internal_attrs.init_from_native_type(field_metadata.data_type)
                field_metadata.description = row["column_comment"]
                field_metadata.uniquekey = row["is_unique"] if row["is_unique"] is not None else False
                field_metadata.is_index = row["is_indexed"]
                field_metadata.notnull = row["is_not_null"]
                self.table_metadata.fields[field_metadata.name] = field_metadata
            if primary_key is None:
                # first field with unicity can be used as primary key
                primary_key = [field_metadata.name for field_metadata in self.table_metadata.fields.values() if field_metadata.uniquekey]
                if len(primary_key) > 0:
                    primary_key = primary_key[0]
                else:
                    primary_key = None
            self.table_metadata.primary_key = primary_key
            self.table_metadata.indexes = [field_metadata.name for field_metadata in self.table_metadata.fields.values() if field_metadata.is_index]
            return self.table_metadata

    def _data_type_map_to_ckan(self, field_metadata:FieldMetadata) -> None:
        """
        Some data types need to be translated
        """
        if field_metadata.harvester_attrs["datatype_keyword"] == "geometry":
            if self.params.ckan_postgis:
                if self.params.ckan_default_target_epsg is not None:
                    # TODO: at this point, the ckan_default_target_epsg does not inherit from ckan
                    geometry_type, geo_epsg = parse_geometry_native_type(field_metadata.data_type)
                    field_metadata.data_type = f"geometry({geometry_type},{self.params.ckan_default_target_epsg})"
                    field_metadata.internal_attrs.init_from_native_type(field_metadata.data_type)
            else:
                field_metadata.data_type = "json"
        return

    def _get_field_query_function(self, field_metadata: FieldMetadata) -> str:
        """
        Force some data types to return as text
        """
        if field_metadata.harvester_attrs["datatype_keyword"] == "geometry":
            if self.params.ckan_postgis:
                return sql_varname_escape(field_metadata.name)  # TODO: test if transfer is successful without converting to a GeoJSON string
            else:
                return f"ST_AsGeoJSON({sql_varname_escape(field_metadata.name)})"
        elif field_metadata.data_type == "jsonb":
            return f"{sql_varname_escape(field_metadata.name)}::text"
        else:
            return sql_varname_escape(field_metadata.name)

    def clean_table_metadata(self) -> TableMetadata:
        table_metadata = self.query_table_metadata().copy()
        for field_metadata in table_metadata.fields.values():
            self._data_type_map_to_ckan(field_metadata)
        return table_metadata

    def update_from_ckan(self, ckan):
        super().update_from_ckan(ckan)
        for field_name, field_metadata in self.table_metadata.fields.items():
            field_metadata.internal_attrs.update_from_ckan(ckan)

    def get_default_primary_key(self) -> List[str]:
        table_metadata = self.query_table_metadata()
        return table_metadata.primary_key

    def _get_sql_fields_query(self):
        return ", ".join([self._get_field_query_function(field_metadata) for field_metadata in self.table_metadata.fields.values()])

    def get_default_data_cleaner(self) -> CkanDataCleanerABC:
        data_cleaner = CkanDataCleanerUploadGeom()
        return data_cleaner

    def list_queries(self, *, new_connection:bool=False) -> List[Tuple[str,int]]:
        self.connect(cancel_if_connected=not new_connection)
        postgre_schema_name = self.params.dataset
        postgre_table_name = self.params.table
        if self.params.verbose_harvester:
            print(f"Counting documents of table {self.params.table}")
        count_query = f"SELECT COUNT(*) FROM {postgre_schema_name}.{postgre_table_name}"
        if self.params.query_string is not None:
            count_query += " " + self.params.query_string
        count_df = pd.read_sql(count_query, self.alchemy_engine)
        num_rows = count_df["count"].iloc[0]
        fields_query = self._get_sql_fields_query()
        request_query_base = f"SELECT {fields_query} FROM {postgre_schema_name}.{postgre_table_name}"
        if self.params.query_string is not None:
            request_query_base += " " + self.params.query_string
        num_queries = num_rows // self.params.limit + 1
        if self.params.single_request:
            return [(f"{request_query_base} LIMIT {self.params.limit} OFFSET {i * self.params.limit}", self.params.limit) for i in range(1)]
        else:
            queries_exact = [(f"{request_query_base} LIMIT {self.params.limit} OFFSET {i * self.params.limit}", self.params.limit) for i in range(num_queries)]
            query_extra = f"{request_query_base} LIMIT {self.params.limit} OFFSET {num_queries * self.params.limit}"
            queries_exact.append((query_extra, self.params.limit))
            return queries_exact

    def query_data(self, query:Dict[str,Any]) -> pd.DataFrame:
        df = pd.read_sql(query, self.alchemy_engine)
        return df




#!python3
# -*- coding: utf-8 -*-
"""
Code to upload metadata to the CKAN server, with one thread per resource
"""
from typing import List, Dict
import threading

from ckanapi_harvesters.auxiliary.ckan_model import CkanState
from ckanapi_harvesters.auxiliary.ckan_progress_callbacks import CkanProgressCallback
from ckanapi_harvesters.builder.builder_package_2_harvesters import BuilderPackageWithHarvesters
from ckanapi_harvesters.builder.builder_resource_multi_abc import BuilderMultiABC
from ckanapi_harvesters.ckan_api import CkanApi


class BuilderPackageMultiThreaded(BuilderPackageWithHarvesters, BuilderMultiABC):
    def __init__(self, package_name: str = None, *, package_id: str = None,
                 title: str = None, description: str = None, private: bool = None, state: CkanState = None,
                 version: str = None,
                 url: str = None, tags: List[str] = None,
                 organization_name: str = None, license_name: str = None):
        super().__init__(package_name=package_name, package_id=package_id,
                         title=title, description=description, private=private, state=state, version=version,
                         url=url, tags=tags, organization_name=organization_name, license_name=license_name)
        # BuilderMultiABC:
        self.progress_callback = CkanProgressCallback()
        self.stop_event = threading.Event()
        self.thread_ckan: Dict[str, CkanApi] = {}
        self.enable_multi_threaded_upload:bool = True
        self.enable_multi_threaded_download:bool = True

    def copy(self, dest=None) -> "BuilderPackageWithHarvesters":
        if dest is None:
            dest = BuilderPackageWithHarvesters()
        super().copy(dest=dest)
        dest.progress_callback = self.progress_callback.copy()
        dest.enable_multi_threaded_upload = self.enable_multi_threaded_upload
        dest.enable_multi_threaded_download = self.enable_multi_threaded_download
        return dest

    # TODO: implement abstract methods





#!python3
# -*- coding: utf-8 -*-
"""
Operations on urls
"""
from typing import Union
import urllib.parse

from ckanapi_harvesters.auxiliary.login import Login


urlsep = '/'


def clean_base_url(url:Union[str,None]) -> Union[str,None]:
    if url is None:
        return None
    elif not url.endswith(urlsep):
        return url + urlsep
    else:
        return url

def url_matches_host(host_url:str, url:str) -> bool:
    # TODO: improve the url matching test
    return url.startswith(host_url)

def is_valid_url(url:str) -> bool:
    try:
        result = urllib.parse.urlparse(url)
        return all([result.scheme, result.netloc, result.path])
    except ValueError as e:
        return False

def url_join(base:str, *args:str) -> str:
    url = base
    for arg in args:
        if len(arg) > 0:
            if not url.endswith(urlsep):
                url += urlsep
            url += arg
    return url

def url_insert_login(url:str, login:Login):
    """
    Insert user authentication parameters in a url
    """
    if login is None:
        return url
    parsed_url = urllib.parse.urlparse(url)
    netloc_with_auth = f"{login.username}:{login.password}@{parsed_url.netloc}"
    updated_url = parsed_url._replace(netloc=netloc_with_auth)
    final_url = urllib.parse.urlunparse(updated_url)
    return final_url



#!python3
# -*- coding: utf-8 -*-
"""
Code to generate sample data for the dataset example
"""
from typing import Tuple
import os
import re

import pandas as pd
import numpy as np

from ckanapi_harvesters.builder.example import example_package_resources_dir


GPS_DIGITS = 6
def degrees_DMS(lat:str, lon:str) -> Tuple[float, float]:
    """
    Returns angles for GPS coordinates in the form
    :param lat: example: 48° 51' 12.24845" N
    :param lon: example: 2° 20' 55.62563" E
    :return:
    """
    deg, minutes, seconds, direction = re.split('[°\'"]', lat)
    lat_val = (float(deg) + float(minutes) / 60 + float(seconds) / (60 * 60)) * (-1 if direction in ['W', 'S'] else 1)
    deg, minutes, seconds, direction = re.split('[°\'"]', lon)
    lon_val = (float(deg) + float(minutes) / 60 + float(seconds) / (60 * 60)) * (-1 if direction in ['W', 'S'] else 1)
    return (lat_val, lon_val)


def run():
    N = 100
    df_users = pd.DataFrame([{"user_id": 1, "age": 30},
                             {"user_id": 2, "age": 80},
                             ])
    point_0 = degrees_DMS('48° 51\' 12.24845" N', '2° 20\' 55.62563" E')
    traces = [{"user_id": 1, "origin": point_0, "destination": degrees_DMS('48°53\'37.6"N', '2°23\'24.9"E')},
              {"user_id": 1, "origin": point_0, "destination": degrees_DMS('48°50\'01.7"N', '2°19\'57.1"E')},
              {"user_id": 2, "origin": point_0, "destination": degrees_DMS('48°52\'45.2"N', '2°18\'32.8"E')},
    ]
    users_file = os.path.join(example_package_resources_dir, "users_local.csv")
    df_users.to_csv(users_file, index=False)
    traces_dir = os.path.join(example_package_resources_dir, "traces")
    traces_dir_multi = os.path.join(example_package_resources_dir, "traces_multi")
    for trace_id, trace in enumerate(traces):
        table_index = np.array([0, N-1])
        table_lat = np.array([trace["origin"][0], trace["destination"][0]])
        table_lon = np.array([trace["origin"][1], trace["destination"][1]])
        df_trace = pd.DataFrame()
        index = np.arange(N)
        xp_index = np.array([0, N-1])
        df_trace["index_in_trace"] = index
        df_trace.insert(loc=0, column="trace_id", value=trace_id)
        df_trace["user_id"] = trace["user_id"]
        df_trace["timestamp"] = pd.Timestamp(year=2025, month=1, day=1, hour=12, minute=0, second=0, microsecond=0) + index.astype('timedelta64[s]')
        df_trace["timestamp_local"] = df_trace["timestamp"] + pd.Timedelta(hours=1)  # local = UTC+1 for winter light saving time in Paris
        df_trace["timestamp"] = df_trace["timestamp"].apply(pd.Timestamp.isoformat)  # ISO-8601 format
        df_trace["timestamp_local"] = df_trace["timestamp_local"].apply(pd.Timestamp.isoformat)  # ISO-8601 format
        df_trace["latitude"] = np.interp(index, xp=xp_index, fp=table_lat)
        df_trace["longitude"] = np.interp(index, xp=xp_index, fp=table_lon)
        df_trace["latitude"] = df_trace["latitude"].values.round(GPS_DIGITS)
        df_trace["longitude"] = df_trace["longitude"].values.round(GPS_DIGITS)
        df_trace.set_index(keys=["trace_id", "index_in_trace"], drop=False, inplace=True)
        assert(df_trace.index.is_unique)  # verify integrity
        trace_file = os.path.join(traces_dir, f"trace_{trace_id:03d}.csv")
        df_trace.to_csv(trace_file, index=False)
        trace_file_multi = os.path.join(traces_dir_multi, f"trace_{trace_id:03d}.csv")
        df_trace.to_csv(trace_file_multi, index=False)
    print("Example data regenerated")


if __name__ == '__main__':
    run()




#!python3
# -*- coding: utf-8 -*-
"""
Auxiliary functions for package upload/download example
"""
from typing import Union, List, Dict

import pandas as pd

from ckanapi_harvesters import CkanField


def users_upload(df_local: Union[pd.DataFrame, List[dict]], *,
                 fields:Dict[str, CkanField], file_name:str=None, total_lines_read:int=None, **kwargs) \
        -> Union[pd.DataFrame, List[dict]]:
    print("<<< Upload function example called on users dataframe containing ids " + ",".join([str(id) for id in df_local["user_id"].to_list()]))
    print(f"<<< File {file_name}")
    return df_local

def users_download(df_download: pd.DataFrame, *,
                   fields:Dict[str, CkanField], file_query:str=None, **kwargs) \
        -> Union[pd.DataFrame, List[dict]]:
    print("<<< Download function example called on users dataframe containing ids " + ",".join([str(id) for id in df_download["user_id"].to_list()]))
    print(f"<<< File query {file_query}")
    return df_download


if __name__ == '__main__':
    df_users = pd.DataFrame({"user_id": [1, 2, 3]})
    df_users = users_upload(df_users, fields={})
    print(df_users)




#!python3
# -*- coding: utf-8 -*-
"""
Data format policy representation and enforcing
"""
from typing import List, Union, Dict
import datetime
from collections import OrderedDict

import pandas as pd

from ckanapi_harvesters.auxiliary.ckan_auxiliary import _string_from_element, _bool_from_string
from ckanapi_harvesters.auxiliary.ckan_defs import ckan_tags_sep
from ckanapi_harvesters.policies.data_format_policy_defs import DataType
from ckanapi_harvesters.policies.data_format_policy_errors import DataPolicyError, ErrorLevel, _policy_msg
from ckanapi_harvesters.policies.data_format_policy_defs import StringMatchMode
from ckanapi_harvesters.policies.data_format_policy_abc import DataPolicyElementABC


class CustomFieldSpecification(DataPolicyElementABC):
    def __init__(self, key: str=None, values: List[str]=None, data_type:DataType=None,
                 match_mode:StringMatchMode=StringMatchMode.Any,
                 help:str=None,
                 mandatory:bool=False, error_level:ErrorLevel=ErrorLevel.Information):
        super().__init__(mandatory=mandatory, error_level=error_level)
        self.key: str = key
        self.values: List[str] = values
        self.data_type:DataType = data_type
        self.match_mode: StringMatchMode = match_mode
        self.help: str = help

    def to_dict(self) -> dict:
        d = OrderedDict([
            ("key", self.key),
            ("values", self.values),
            ("data_type", self.data_type.name if self.data_type is not None else ""),
            ("match_mode", self.match_mode.name),
             ])
        d.update(super().to_dict())
        return d

    @staticmethod
    def from_dict(d:dict) -> "CustomFieldSpecification":
        obj = CustomFieldSpecification()
        obj._load_from_dict(d)
        return obj

    def _load_from_dict(self, d:dict):
        super()._load_from_dict(d)
        self.key = d["key"]
        self.values = d["values"] if "values" in d.keys() else None
        self.data_type = DataType.from_str(d["data_type"]) if "data_type" in d.keys() and not d["data_type"] == "" else None
        self.match_mode = StringMatchMode.from_str(d["match_mode"]) if "match_mode" in d.keys() else None

    @staticmethod
    def from_df_row(row: pd.Series) -> "CustomFieldSpecification":
        key = _string_from_element(row["key"]).strip()
        values_str = _string_from_element(row["values"])
        values = values_str.split(ckan_tags_sep)
        mode_str = _string_from_element(row["mode"])
        mode = StringMatchMode.from_str(mode_str) if mode_str is not None else StringMatchMode.Any
        help:Union[str,None] = None
        if "help" in row.keys():
            help = _string_from_element(row["help"])
        return CustomFieldSpecification(key=key, values=values, match_mode=mode, help=help)

    def enforce(self, values: str, *, context:str=None, verbose: bool = True, buffer:List[DataPolicyError]=None) -> bool:
        key_context = context + " / custom key " + self.key
        value = values
        specs = self.values
        if specs is None:
            return not self.match_mode == StringMatchMode.NotEmpty
        if self.data_type is None or self.data_type == DataType.Text:
            success = self._enforce_unit_string(value, specs, context=key_context, verbose=verbose, buffer=buffer)
        elif self.data_type == DataType.Bool:
            self.match_mode = StringMatchMode.Match
            success = self._enforce_unit_string(value, {"True", "False"}, context=key_context, verbose=verbose, buffer=buffer)
        elif self.data_type == DataType.TimeStamp:
            if value is not None and len(value) > 0:
                try:
                    timestamp = datetime.datetime.fromisoformat(value)
                except Exception as e:
                    return False
                success = True
            else:
                success = True
        elif self.data_type == DataType.Numeric:
            self.match_mode = StringMatchMode.Regex
            success = self._enforce_unit_string(value, "/d+", context=key_context, verbose=verbose, buffer=buffer)
        else:
            raise NotImplementedError("Unsupported data type: " + str(self.data_type))
        return success


class CustomFieldsPolicy(DataPolicyElementABC):
    def __init__(self, custom_fields_spec:List[CustomFieldSpecification]=None,
                 restrict_to_list:ErrorLevel=ErrorLevel.Information, keys_case_sensitive:bool=True,
                 mandatory:bool=False, error_level:ErrorLevel=ErrorLevel.Information):
        super().__init__(mandatory=mandatory, error_level=error_level)
        if custom_fields_spec is None:
            custom_fields_spec = []
        self.restrict_to_list: ErrorLevel = restrict_to_list
        self.keys_case_sensitive:bool = keys_case_sensitive
        self.custom_fields_spec:Dict[str,CustomFieldSpecification] = {}
        if keys_case_sensitive:
            self.custom_fields_spec = {keypair_spec.key: keypair_spec for keypair_spec in custom_fields_spec}
        else:
            self.custom_fields_spec = {keypair_spec.key.lower(): keypair_spec for keypair_spec in custom_fields_spec}

    def to_dict(self) -> dict:
        d = {"custom_fields": [spec.to_dict() for spec in self.custom_fields_spec.values()],
             "keys_case_sensitive": self.keys_case_sensitive, "restrict_to_list": self.restrict_to_list.name}
        d.update(super().to_dict())
        return d

    @staticmethod
    def from_dict(d:dict) -> "CustomFieldsPolicy":
        obj = CustomFieldsPolicy()
        obj._load_from_dict(d)
        return obj

    def _load_from_dict(self, d:dict):
        super()._load_from_dict(d)
        self.custom_fields_spec = {spec["key"]: CustomFieldSpecification.from_dict(spec) for spec in d["custom_fields"]}
        self.keys_case_sensitive = _bool_from_string(d["keys_case_sensitive"]) if "keys_case_sensitive" in d.keys() else None
        self.restrict_to_list = ErrorLevel.from_str(d["restrict_to_list"]) if "restrict_to_list" in d.keys() else None

    def enforce(self, values: Dict[str, str], *, context:str=None, verbose: bool = True, buffer:List[DataPolicyError]=None) -> bool:
        success = True
        if self.keys_case_sensitive:
            keys = set(values.keys())
        else:
            keys = {key.lower() for key in values.keys()}
        extra_keys = keys - set(self.custom_fields_spec.keys())
        if len(extra_keys) > 0:
            msg = DataPolicyError(context, self.restrict_to_list, f"Custom keys do not make part of the defined list: {','.join(extra_keys)}")
            _policy_msg(msg, error_level=self.error_level, buffer=buffer, verbose=verbose)
            success = False
        mandatory_keys = {key for key, keypair in self.custom_fields_spec.items() if keypair.mandatory}
        missing_keys = mandatory_keys - keys
        if len(missing_keys) > 0:
            msg = DataPolicyError(context, self.restrict_to_list, f"Mandatory custom keys were not found: {', '.join(missing_keys)}")
            _policy_msg(msg, error_level=self.error_level, buffer=buffer, verbose=verbose)
            success = False
        for key, value in values.items():
            key_context = context + " / custom key " + key
            if not self.keys_case_sensitive:
                key = key.lower()
            spec = self.custom_fields_spec[key] if key in self.custom_fields_spec.keys() else None
            if spec is not None:
                success_value = spec.enforce(value, context=context, verbose=verbose, buffer=buffer)
                success &= success_value
        return success





#!python3
# -*- coding: utf-8 -*-
"""
File format keyword selection
"""
from typing import Union

from ckanapi_harvesters.harvesters.data_cleaner.data_cleaner_abc import CkanDataCleanerABC, DataCleanerNone
from ckanapi_harvesters.harvesters.data_cleaner.data_cleaner_upload import CkanDataCleanerUpload
from ckanapi_harvesters.harvesters.data_cleaner.data_cleaner_upload_1_basic import CkanDataCleanerUploadBasic, CkanDataCleanerUploadDigitalColumns
from ckanapi_harvesters.harvesters.data_cleaner.data_cleaner_upload_2_geom import CkanDataCleanerUploadGeom
from ckanapi_harvesters.harvesters.data_cleaner.data_cleaner_upload_3_check import CkanDataCleanerUploadCheckOnly

data_cleaner_dict = {
    "all": CkanDataCleanerUpload,
    "basic": CkanDataCleanerUploadBasic,
    "digitalcolumns": CkanDataCleanerUploadDigitalColumns,
    "geojson": CkanDataCleanerUploadGeom,
    "check": CkanDataCleanerUploadCheckOnly,
    "none": DataCleanerNone,
    # NB: member "mongo" is added when loading ckanapi_harvesters.harvesters.pymongo_data_cleaner
}

def init_data_cleaner(data_cleaner_string:Union[str,None]) -> Union[CkanDataCleanerABC,None]:
    if data_cleaner_string is None or data_cleaner_string == "":
        return None
    data_cleaner_name = data_cleaner_string.lower().strip()
    if data_cleaner_name in data_cleaner_dict.keys():
        data_cleaner_class = data_cleaner_dict[data_cleaner_name]
        return data_cleaner_class()
    else:
        raise NotImplementedError(f'Data cleaner {data_cleaner_name} not implemented')


